# -*- coding: utf-8 -*-
import numpy as np
import pytest
from aspired import util

dummy_image = np.ones((1000, 1000))
dummy_image_nan = np.ones((1000, 1000))
dummy_image_none = np.ones((1000, 1000))
for i in range(100):
    dummy_image[i, i] = 65000
    dummy_image_nan[i, i] = np.nan
    dummy_image_none[i, i] = None

dummy_sum = np.sum(dummy_image)


# Healing the diagonal of pixels with 65000 count
def test_cutoff_mask():
    cutoff_mask, _ = util.create_cutoff_mask(dummy_image, cutoff=60000)
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Healing the diagonal of pixels with 65000 count
def test_cutoff_mask_list():
    cutoff_mask, _ = util.create_cutoff_mask(dummy_image, cutoff=[0, 1000])
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Should do nothing here
def test_cutoff_mask_does_nothing():
    cutoff_mask, _ = util.create_cutoff_mask(dummy_image, cutoff=65500)
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == dummy_sum


def test_cutoff_mask_1D_array():
    util.create_cutoff_mask(np.ones(100))


@pytest.mark.xfail()
def test_cutoff_mask_expect_fail_list():
    util.create_cutoff_mask(dummy_image, cutoff=[0, 1000, 60000])


@pytest.mark.xfail()
def test_cutoff_mask_expect_fail_str():
    util.create_cutoff_mask(dummy_image, cutoff="10")


# Repeat with grow set to True
# Healing the diagonal of pixels with 65000 count
def test_cutoff_mask_grow():
    cutoff_mask, _ = util.create_cutoff_mask(
        dummy_image, cutoff=60000, grow=True
    )
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Healing the diagonal of pixels with 65000 count
def test_cutoff_mask_list_grow():
    cutoff_mask, _ = util.create_cutoff_mask(
        dummy_image, cutoff=[0, 1000], grow=True
    )
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Should do nothing here
def test_cutoff_mask_does_nothing_grow():
    cutoff_mask, _ = util.create_cutoff_mask(
        dummy_image, cutoff=65500, grow=True
    )
    healed_image = util.bfixpix(dummy_image, cutoff_mask, retdat=True)
    assert np.sum(healed_image) == dummy_sum


# Test bad mask
# Healing the diagonal of pixels with NAN
def test_bad_mask():
    bad_mask, _ = util.create_bad_pixel_mask(dummy_image_nan)
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image_nan)


# Healing the diagonal of pixels with None
def test_bad_mask_list():
    bad_mask, _ = util.create_bad_pixel_mask(dummy_image_none)
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image_none)


# Should do nothing here
def test_bad_mask_does_nothing():
    bad_mask, _ = util.create_bad_pixel_mask(dummy_image)
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == dummy_sum


def test_bad_mask_1D_array():
    util.create_bad_pixel_mask(np.ones(100))


# Repeat with grow set to True
# Healing the diagonal of pixels with NAN
def test_bad_mask_grow():
    bad_mask, _ = util.create_bad_pixel_mask(
        dummy_image_nan, grow=True, diagonal=True
    )
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Healing the diagonal of pixels with None
def test_bad_mask_list_grow():
    bad_mask, _ = util.create_bad_pixel_mask(
        dummy_image_none, grow=True, diagonal=True
    )
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == np.size(dummy_image)


# Should do nothing here
def test_bad_mask_does_nothing_grow():
    bad_mask, _ = util.create_bad_pixel_mask(
        dummy_image, grow=True, diagonal=True
    )
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=True)
    assert np.sum(healed_image) == dummy_sum


# Should return nothing here
def test_bad_mask_does_nothing_grow_none():
    bad_mask, _ = util.create_bad_pixel_mask(
        dummy_image, grow=True, diagonal=True
    )
    healed_image = util.bfixpix(dummy_image, bad_mask, retdat=False)
    assert np.sum(healed_image) is None



# -*- coding: utf-8 -*-
import os

import numpy as np
from aspired import image_reduction, spectral_reduction
from astropy.io import fits

HERE = os.path.dirname(os.path.realpath(__file__))


def gaussian(pixels, central_pixel):
    """
    Parameters
    ----------
    log_age: array
        the age to return the SFH.
    peak_age: float
        the time of the maximum star formation.
    Returns
    -------
    The relative SFH at the given log_age location.
    """
    stdv = 1.0
    variance = stdv**2.0
    g = (
        np.exp(-((pixels - central_pixel) ** 2.0) / 2 / variance)
        / np.sqrt(2 * np.pi)
        / stdv
    )
    return g


# background noise
bg_level = 5.0

# Prepare dummy data
# total signal should be [2 * (2 + 5 + 10) + 20] - [1 * 7] = 47
dummy_data = np.ones((100, 1000)) * bg_level

dummy_data[47] += 2.0
dummy_data[48] += 5.0
dummy_data[49] += 10.0
dummy_data[50] += 50.0
dummy_data[51] += 10.0
dummy_data[52] += 5.0
dummy_data[53] += 2.0
dummy_data = np.random.normal(dummy_data)


# Prepare dummy gaussian data
dummy_gaussian_data = (
    np.ones((100, 1000)).T * gaussian(np.arange(100), 50)
).T * 10000.0
dummy_gaussian_data = (
    np.random.normal(dummy_gaussian_data, scale=bg_level) + bg_level
)

# Prepare faint dummy gaussian data
dummy_gaussian_data_faint = (
    np.ones((100, 1000)).T * gaussian(np.arange(100), 50)
).T * 100.0
dummy_gaussian_data_faint = (
    np.random.normal(dummy_gaussian_data_faint, scale=bg_level) + bg_level
)


def test_spectral_extraction():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_data,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="DEBUG",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        ap_faint=0,
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(HERE, "test_output", "test_extraction_aptrace"),
        return_jsonstring=True,
    )
    trace = np.round(np.mean(dummy_twodspec.spectrum_list[0].trace))
    assert np.isclose(trace, 35, atol=1.0), (
        "Trace is at row "
        + str(trace)
        + ", but it is expected to be at row 35."
    )

    # Optimal extracting spectrum by summing over the aperture along the trace
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=False,
        filename=os.path.join(
            HERE, "test_output", "test_extraction_apextract"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )

    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    assert np.isclose(count, 84, atol=1.0), (
        "Extracted count is " + str(count) + " but it should be 84."
    )

    dummy_twodspec.inspect_extracted_spectrum(
        display=False,
        filename=os.path.join(
            HERE, "test_output", "test_extraction_extracted_spectrum"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )


def test_gaussian_spectral_extraction():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )
    trace = np.round(np.mean(dummy_twodspec.spectrum_list[0].trace))
    assert np.isclose(trace, 35, atol=1.0), (
        "Trace is at row "
        + str(trace)
        + ", but it is expected to be at row 35."
    )

    # Direct extraction by summing over the aperture along the trace
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=False,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_tophat = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Horne86 gauss)
    dummy_twodspec.ap_extract(apwidth=5, optimal=True, model="gauss")
    dummy_twodspec.inspect_line_spread_function(
        display=False,
        filename=os.path.join(
            HERE, "test_output", "test_extraction_line_spread_function_gauss"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Horne86 lowess)
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=True,
        model="lowess",
        lowess_frac=0.05,
    )
    dummy_twodspec.inspect_line_spread_function(
        display=False,
        filename=os.path.join(
            HERE, "test_output", "test_extraction_line_spread_function_lowess"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    print(count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Marsh89)
    dummy_twodspec.ap_extract(apwidth=5, optimal=True, algorithm="marsh89")
    dummy_twodspec.inspect_line_spread_function(
        display=False,
        filename=os.path.join(
            HERE, "test_output", "test_extraction_line_spread_function_marsh"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_marsh = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )


def test_gaussian_spectral_extraction_top_hat_low_signal():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data_faint,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )
    trace = np.round(np.mean(dummy_twodspec.spectrum_list[0].trace))
    assert np.isclose(trace, 35, atol=1.0), (
        "Trace is at row "
        + str(trace)
        + ", but it is expected to be at row 35."
    )

    # Direct extraction by summing over the aperture along the trace
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=False,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_tophat = count / count_err
    assert np.isclose(count, 100.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~100."
    )


def test_gaussian_spectral_extraction_horne86_gaussian_low_signal():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data_faint,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )

    # Optimal extraction (Horne86 gauss)
    dummy_twodspec.ap_extract(
        apwidth=5, optimal=True, algorithm="horne86", model="gauss"
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 100.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~100."
    )


def test_gaussian_spectral_extraction_horne86_lowess_low_signal():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data_faint,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )

    # Optimal extraction (Horne86 lowess)
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=True,
        model="lowess",
        lowess_frac=0.05,
    )
    count = np.median(dummy_twodspec.spectrum_list[0].count)
    count_err = np.median(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 100.0, rtol=0.01, atol=count_err * 2.0), (
        "Extracted count is " + str(count) + " but it should be ~100."
    )


def test_gaussian_spectral_extraction_marsh89_low_signal():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data_faint,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=1.0,
        seeing=1.0,
        exptime=1.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )

    # Optimal extraction (Marsh89)
    dummy_twodspec.ap_extract(apwidth=5, optimal=True, algorithm="marsh89")
    count = np.median(dummy_twodspec.spectrum_list[0].count)
    count_err = np.median(dummy_twodspec.spectrum_list[0].count_err)
    snr_marsh = count / count_err
    assert np.isclose(count, 100.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~100."
    )


def test_user_supplied_trace():
    spatial_mask = np.arange(20, 200)
    spec_mask = np.arange(100, 1024)

    # Loading a single pre-saved spectral trace.
    lhs6328_extracted = fits.open(
        os.path.join(HERE, "test_data", "test_full_run_science_0.fits")
    )
    lhs6328_trace = lhs6328_extracted["trace"].data
    lhs6328_trace_sigma = lhs6328_extracted["trace_sigma"].data

    lhs6328_twodspec = spectral_reduction.TwoDSpec(
        os.path.join(HERE, "test_data", "v_e_20180810_12_1_0_0.fits.gz"),
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        readnoise=2.34,
        log_file_name=None,
    )

    lhs6328_twodspec.add_trace(
        trace=lhs6328_trace, trace_sigma=lhs6328_trace_sigma
    )

    lhs6328_twodspec.ap_extract(
        apwidth=15, optimal=True, skywidth=10, skydeg=1, display=False
    )

    lhs6328_twodspec.save_fits(
        output="count",
        filename=os.path.join(
            HERE, "test_output", "user_supplied_trace_for_extraction"
        ),
        overwrite=True,
    )

    lhs6328_twodspec.remove_trace()


spatial_mask = np.arange(30, 200)
spec_mask = np.arange(50, 1024)

# Science frame
lhs6328_frame = image_reduction.ImageReduction(
    log_level="INFO", log_file_name=None
)
lhs6328_frame.add_filelist(
    os.path.join(HERE, "test_data", "sprat_LHS6328.list")
)
lhs6328_frame.load_data()
lhs6328_frame.reduce()

lhs6328_twodspec = spectral_reduction.TwoDSpec(
    lhs6328_frame,
    spatial_mask=spatial_mask,
    spec_mask=spec_mask,
    cosmicray=True,
    readnoise=5.7,
    log_level="DEBUG",
    log_file_name=None,
)

lhs6328_twodspec.ap_trace(nspec=1, display=False)


def test_tophat_extraction():
    lhs6328_twodspec.ap_extract(optimal=False, display=False)
    lhs6328_twodspec.ap_extract(optimal=False, display=False, spec_id=0)
    lhs6328_twodspec.ap_extract(optimal=False, display=False, spec_id=[0])
    lhs6328_twodspec.ap_extract(
        optimal=False,
        apwidth=[4, 5],
        skysep=[8, 10],
        skywidth=[3, 4],
        display=False,
    )


def test_extraction_wrong_size_extraction_description():
    lhs6328_twodspec.ap_extract(
        optimal=False, apwidth=[1, 2, 3], skysep=[4, 5, 6], skywidth=[7, 8, 9]
    )


def test_horne_extraction():
    lhs6328_twodspec.ap_extract(
        optimal=True, algorithm="horne86", display=False
    )


def test_marsh_extraction_fast():
    lhs6328_twodspec.ap_extract(
        optimal=True, algorithm="marsh89", qmode="fast-nearest", display=False
    )


def test_marsh_extraction_fast_linear():
    lhs6328_twodspec.ap_extract(
        optimal=True, algorithm="marsh89", qmode="fast-linear", display=False
    )


def test_marsh_extraction_slow():
    lhs6328_twodspec.ap_extract(
        optimal=True, algorithm="marsh89", qmode="slow-nearest", display=False
    )


def test_marsh_extraction_slow_linear():
    lhs6328_twodspec.ap_extract(
        optimal=True, algorithm="marsh89", qmode="slow-linear", display=False
    )


def test_marsh_extraction_fast_apwidth():
    lhs6328_twodspec.ap_extract(
        optimal=True,
        apwidth=5,
        algorithm="marsh89",
        qmode="fast-nearest",
        display=False,
    )


def test_marsh_extraction_fast_apwidth_str():
    lhs6328_twodspec.ap_extract(
        optimal=True,
        apwidth="blabla",
        algorithm="marsh89",
        qmode="fast-nearest",
        display=False,
    )


# Exposure time and gain should not have an effect on the extraction itself
def test_gaussian_spectral_extraction_10000s_exptime_2x_gain():
    # masking
    spec_mask = np.arange(10, 900)
    spatial_mask = np.arange(15, 85)

    # initialise the two spectral_reduction.TwoDSpec()
    dummy_twodspec = spectral_reduction.TwoDSpec(
        dummy_gaussian_data,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        log_file_name=None,
        log_level="CRITICAL",
        saxis=1,
        flip=False,
        cosmicray_sigma=5.0,
        readnoise=0.1,
        gain=2.0,
        seeing=1.0,
        exptime=10000.0,
    )

    # Trace the spectrum, note that the first 15 rows were trimmed from the
    # spatial_mask
    dummy_twodspec.ap_trace(
        rescale=True,
        fit_deg=0,
    )
    trace = np.round(np.mean(dummy_twodspec.spectrum_list[0].trace))
    assert np.isclose(trace, 35, atol=1.0), (
        "Trace is at row "
        + str(trace)
        + ", but it is expected to be at row 35."
    )

    # Direct extraction by summing over the aperture along the trace
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=False,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_tophat = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Horne86 gauss)
    dummy_twodspec.ap_extract(apwidth=5, optimal=True, model="gauss")
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Horne86 lowess)
    dummy_twodspec.ap_extract(
        apwidth=5,
        optimal=True,
        model="lowess",
        lowess_frac=0.05,
    )
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_horne = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )

    # Optimal extraction (Marsh89)
    dummy_twodspec.ap_extract(apwidth=5, optimal=True, algorithm="marsh89")
    count = np.mean(dummy_twodspec.spectrum_list[0].count)
    count_err = np.mean(dummy_twodspec.spectrum_list[0].count_err)
    snr_marsh = count / count_err
    assert np.isclose(count, 10000.0, rtol=0.01, atol=count_err), (
        "Extracted count is " + str(count) + " but it should be ~10000."
    )



from aspired import standard_list
from astroquery.simbad import Simbad

result_table = Simbad.query_objectids("hr9087")
for i in result_table["ID"].value:
    print('"{}",'.format(" ".join(str(i.decode("ascii")).split())))



# -*- coding: utf-8 -*-
import copy
import os
from unittest.mock import patch

import numpy as np
import pytest
from aspired import spectral_reduction
from aspired.spectrum_oneD import SpectrumOneD
from aspired.wavelength_calibration import WavelengthCalibration
from astropy.io import fits

HERE = os.path.dirname(os.path.realpath(__file__))

# Line list
atlas = [
    4193.5,
    4385.77,
    4500.98,
    4524.68,
    4582.75,
    4624.28,
    4671.23,
    4697.02,
    4734.15,
    4807.02,
    4921.48,
    5028.28,
    5618.88,
    5823.89,
    5893.29,
    5934.17,
    6182.42,
    6318.06,
    6472.841,
    6595.56,
    6668.92,
    6728.01,
    6827.32,
    6976.18,
    7119.60,
    7257.9,
    7393.8,
    7584.68,
    7642.02,
    7740.31,
    7802.65,
    7887.40,
    7967.34,
    8057.258,
]
element = ["Xe"] * len(atlas)

arc_spec = np.loadtxt(
    os.path.join(HERE, "test_data", "test_full_run_science_0_arc_spec.csv"),
    delimiter=",",
    skiprows=1,
)

wavecal = WavelengthCalibration(log_file_name=None)
wavecal.add_arc_spec(arc_spec)

# Find the peaks of the arc
wavecal.find_arc_lines()

arc_lines = wavecal.spectrum_oned.peaks

np.random.seed(0)


def test_wavecal():
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal = WavelengthCalibration(log_file_name=None)

    # Science arc_spec
    lhs6328_spectrum_oned.add_arc_spec(arc_spec)
    wavecal.from_spectrum_oned(lhs6328_spectrum_oned)

    # Find the peaks of the arc
    wavecal.find_arc_lines(
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_find_arc_lines"
        ),
        display=False,
        return_jsonstring=True,
    )

    # Configure the wavelength calibrator
    wavecal.initialise_calibrator()
    wavecal.set_hough_properties(
        num_slopes=1000,
        xbins=200,
        ybins=200,
        min_wavelength=3500,
        max_wavelength=8500,
    )
    wavecal.set_ransac_properties(filter_close=True)

    wavecal.add_user_atlas(elements=element, wavelengths=atlas)

    # Remove all lines between 3500 and 4000
    wavecal.remove_atlas_lines_range(wavelength=3750, tolerance=250)
    wavecal.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    wavecal.fit(max_tries=500, display=False)

    # Getting the calibrator
    wavecal.get_calibrator()

    # Save a FITS file
    wavecal.save_fits(
        output="wavecal_coefficients",
        filename=os.path.join(HERE, "test_output", "test_wavecal"),
        overwrite=True,
    )

    # Save a CSV file
    wavecal.save_csv(
        output="wavecal_coefficients",
        filename=os.path.join(HERE, "test_output", "test_wavecal"),
        overwrite=True,
    )

    # Getting the calibrator
    wavecal.get_spectrum_oned()

    wavecal.list_atlas()
    wavecal.clear_atlas()
    wavecal.list_atlas()


def test_setting_a_known_pair():
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal = WavelengthCalibration(log_file_name=None)
    lhs6328_spectrum_oned.add_arc_spec(arc_spec)
    wavecal.from_spectrum_oned(lhs6328_spectrum_oned)
    # Find the peaks of the arc
    wavecal.find_arc_lines(
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_find_arc_lines"
        ),
        display=False,
        return_jsonstring=True,
    )
    wavecal.initialise_calibrator()
    wavecal.set_known_pairs(123, 456)
    assert wavecal.spectrum_oned.calibrator.pix_known == 123
    assert wavecal.spectrum_oned.calibrator.wave_known == 456


@patch("plotly.graph_objects.Figure.show")
def test_setting_known_pairs(mock_show):
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal = WavelengthCalibration(log_file_name=None)
    lhs6328_spectrum_oned.add_arc_spec(arc_spec)
    wavecal.from_spectrum_oned(lhs6328_spectrum_oned)
    # Find the peaks of the arc
    wavecal.find_arc_lines(
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_find_arc_lines"
        ),
        display=True,
        return_jsonstring=True,
    )
    wavecal.initialise_calibrator()
    wavecal.set_known_pairs([123, 234], [456, 567])
    assert len(wavecal.spectrum_oned.calibrator.pix_known) == 2
    assert len(wavecal.spectrum_oned.calibrator.wave_known) == 2


@pytest.mark.xfail()
def test_setting_a_none_to_known_pairs_expect_fail():
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal = WavelengthCalibration(log_file_name=None)
    lhs6328_spectrum_oned.add_arc_spec(arc_spec)
    wavecal.from_spectrum_oned(lhs6328_spectrum_oned)
    # Find the peaks of the arc
    wavecal.find_arc_lines(
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_find_arc_lines"
        ),
        display=False,
        return_jsonstring=True,
    )
    wavecal.initialise_calibrator()
    wavecal.set_known_pairs([1.0], [None])


@pytest.mark.xfail()
def test_setting_nones_to_known_pairs_expect_fail():
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal = WavelengthCalibration(log_file_name=None)
    lhs6328_spectrum_oned.add_arc_spec(arc_spec)
    wavecal.from_spectrum_oned(lhs6328_spectrum_oned)
    # Find the peaks of the arc
    wavecal.find_arc_lines(
        save_fig=True,
        fig_type="iframe+png",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_find_arc_lines"
        ),
        display=False,
        return_jsonstring=True,
    )
    wavecal.initialise_calibrator()
    wavecal.set_known_pairs([None], [None])


def test_user_supplied_arc_spec():
    wavecal = WavelengthCalibration(log_file_name=None)

    # Science arc_spec
    wavecal.add_arc_spec(arc_spec)

    # Find the peaks of the arc
    wavecal.find_arc_lines()

    # Configure the wavelength calibrator
    wavecal.initialise_calibrator()
    wavecal.set_hough_properties(
        num_slopes=200,
        xbins=40,
        ybins=40,
        min_wavelength=3500,
        max_wavelength=8500,
    )
    wavecal.set_ransac_properties(filter_close=True)

    wavecal.add_user_atlas(elements=element, wavelengths=atlas)
    wavecal.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    wavecal.fit(max_tries=500, display=False)

    # Save a FITS file
    wavecal.save_fits(
        output="wavecal",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_user_supplied_arc_spec"
        ),
        overwrite=True,
    )

    wavecal.remove_arc_lines()


def test_user_supplied_arc_spec_2():
    wavecal = WavelengthCalibration(log_file_name=None)

    # Find the peaks of the arc
    wavecal.find_arc_lines(arc_spec=arc_spec)

    # Configure the wavelength calibrator
    wavecal.initialise_calibrator()
    wavecal.set_hough_properties(
        num_slopes=200,
        xbins=40,
        ybins=40,
        min_wavelength=3500,
        max_wavelength=8500,
    )
    wavecal.set_ransac_properties(filter_close=True)

    wavecal.add_user_atlas(elements=element, wavelengths=atlas)
    wavecal.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    wavecal.fit(max_tries=500, display=False)

    # Save a FITS file
    wavecal.save_fits(
        output="wavecal",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_user_supplied_arc_spec"
        ),
        overwrite=True,
    )

    wavecal.remove_arc_lines()


def test_user_supplied_arc_spec_arc_lines_from_at_initilisation():
    wavecal = WavelengthCalibration(log_file_name=None)

    # Configure the wavelength calibrator
    wavecal.initialise_calibrator(peaks=arc_lines, arc_spec=arc_spec)
    wavecal.set_hough_properties(
        num_slopes=200,
        xbins=40,
        ybins=40,
        min_wavelength=3500,
        max_wavelength=8500,
    )
    wavecal.set_ransac_properties(filter_close=True)

    wavecal.add_user_atlas(elements=element, wavelengths=atlas)
    wavecal.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    wavecal.fit(max_tries=500, display=False)

    # Save a FITS file
    wavecal.save_fits(
        output="wavecal",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_user_supplied_arc_spec"
        ),
        overwrite=True,
    )


def test_overwritten_copy_of_spectrum_oneds_are_different():
    lhs6328_spectrum_oned = SpectrumOneD(log_file_name=None)
    wavecal_1 = WavelengthCalibration(log_file_name=None)
    wavecal_1.from_spectrum_oned(lhs6328_spectrum_oned)
    memory_1 = id(wavecal_1.spectrum_oned)
    wavecal_1.from_spectrum_oned(
        copy.copy(lhs6328_spectrum_oned), overwrite=True
    )
    memory_2 = id(wavecal_1.spectrum_oned)

    assert memory_1 != memory_2


@pytest.mark.skip(reason="waiting on rascal fix.")
@patch("plotly.graph_objects.Figure.show")
def test_user_supplied_arc_lines(mock_show):
    wavecal = WavelengthCalibration(log_file_name=None)

    # Find the peaks of the arc
    wavecal.add_arc_lines(arc_lines)

    # Configure the wavelength calibrator
    wavecal.initialise_calibrator()
    wavecal.set_hough_properties(
        num_slopes=200,
        xbins=40,
        ybins=40,
        min_wavelength=3500,
        max_wavelength=8500,
    )
    wavecal.set_ransac_properties(filter_close=True)

    wavecal.add_user_atlas(elements=element, wavelengths=atlas)
    wavecal.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    wavecal.fit(max_tries=500, display=True)

    # Save a FITS file
    wavecal.save_fits(
        output="wavecal",
        filename=os.path.join(
            HERE, "test_output", "test_wavecal_user_supplied_arc_spec"
        ),
        overwrite=True,
    )

    wavecal.remove_arc_lines()


def test_user_supplied_poly_coeff_twodspec():
    # Load the image
    lhs6328_fits = fits.open(
        os.path.join(HERE, "test_data", "v_e_20180810_12_1_0_0.fits.gz")
    )[0]
    spatial_mask = np.arange(50, 200)
    spec_mask = np.arange(50, 1024)

    #
    # Loading two pre-saved spectral traces from a single FITS file.
    #
    lhs6328 = spectral_reduction.TwoDSpec(
        lhs6328_fits,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        cosmicray=True,
        readnoise=2.34,
        log_file_name=None,
    )

    # Trace the spectra
    lhs6328.ap_trace(nspec=2, display=False)

    # Extract the spectra
    lhs6328.ap_extract(apwidth=10, optimal=True, skywidth=10, display=False)

    # Calibrate the 1D spectra
    lhs6328_onedspec = spectral_reduction.OneDSpec(log_file_name=None)
    lhs6328_onedspec.from_twodspec(lhs6328)

    fit_coeff = np.array(
        [
            3.09833375e03,
            5.98842823e00,
            -2.83963934e-03,
            2.84842392e-06,
            -1.03725267e-09,
        ]
    )
    fit_type = "poly"

    # Note that there are two science traces, so two polyfit coefficients
    # have to be supplied by in a list
    lhs6328_onedspec.add_fit_coeff(fit_coeff, fit_type)
    lhs6328_onedspec.apply_wavelength_calibration()

    # Inspect reduced spectrum
    lhs6328_onedspec.inspect_reduced_spectrum(display=False)

    # Save as a FITS file
    lhs6328_onedspec.save_fits(
        output="wavecal_coefficients+count",
        filename=os.path.join(
            HERE,
            "test_output",
            "user_supplied_wavelength_polyfit_coefficients",
        ),
        stype="science",
        overwrite=True,
    )


def test_user_supplied_poly_coeff_and_add_arc_twodspec():
    # Load the image
    lhs6328_fits = fits.open(
        os.path.join(HERE, "test_data", "v_e_20180810_12_1_0_0.fits.gz")
    )[0]
    spatial_mask = np.arange(50, 200)
    spec_mask = np.arange(50, 1024)

    #
    # Loading two pre-saved spectral traces from a single FITS file.
    #
    lhs6328 = spectral_reduction.TwoDSpec(
        lhs6328_fits,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        cosmicray=True,
        readnoise=2.34,
        log_file_name=None,
    )

    # Trace the spectra
    lhs6328.ap_trace(nspec=2, display=False)

    # Extract the spectra
    lhs6328.ap_extract(apwidth=10, optimal=True, skywidth=10, display=False)

    # Supply arc manually
    lhs6328.add_arc(
        os.path.join(HERE, "test_data", "v_a_20180810_13_1_0_1.fits.gz")
    )
    lhs6328.apply_mask_to_arc()

    # Calibrate the 1D spectra
    lhs6328_onedspec = spectral_reduction.OneDSpec(log_file_name=None)
    lhs6328_onedspec.from_twodspec(lhs6328)

    fit_coeff = np.array(
        [
            3.09833375e03,
            5.98842823e00,
            -2.83963934e-03,
            2.84842392e-06,
            -1.03725267e-09,
        ]
    )
    fit_type = "poly"

    # Note that there are two science traces, so two polyfit coefficients
    # have to be supplied by in a list
    lhs6328_onedspec.add_fit_coeff(fit_coeff, fit_type)
    lhs6328_onedspec.apply_wavelength_calibration()

    # Inspect reduced spectrum
    lhs6328_onedspec.inspect_reduced_spectrum(display=False)

    # Save as a FITS file
    lhs6328_onedspec.save_fits(
        output="wavecal_coefficients+count",
        filename=os.path.join(
            HERE,
            "test_output",
            "user_supplied_wavelength_polyfit_coefficients",
        ),
        stype="science",
        overwrite=True,
    )


@patch("plotly.graph_objects.Figure.show")
def test_user_supplied_wavelength_twodspec(mock_show):
    # Load the image
    lhs6328_fits = fits.open(
        os.path.join(HERE, "test_data", "v_e_20180810_12_1_0_0.fits.gz")
    )[0]
    spatial_mask = np.arange(50, 200)
    spec_mask = np.arange(50, 1024)

    #
    # Loading two pre-saved spectral traces from a single FITS file.
    #
    lhs6328 = spectral_reduction.TwoDSpec(
        lhs6328_fits,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        cosmicray=True,
        readnoise=2.34,
        log_file_name=None,
    )

    # Trace the spectra
    lhs6328.ap_trace(nspec=2, display=True)

    # Extract the spectra
    lhs6328.ap_extract(apwidth=10, optimal=True, skywidth=10, display=True)

    # Calibrate the 1D spectra
    lhs6328_onedspec = spectral_reduction.OneDSpec(log_file_name=None)
    lhs6328_onedspec.from_twodspec(lhs6328)

    wavelength = np.genfromtxt(
        os.path.join(
            HERE, "test_data", "test_full_run_standard_wavelength.csv"
        )
    )
    # Manually supply wavelengths
    lhs6328_onedspec.add_wavelength([wavelength, wavelength])

    # Inspect reduced spectrum
    lhs6328_onedspec.inspect_reduced_spectrum(display=False)

    # Save as a FITS file
    lhs6328_onedspec.save_fits(
        output="count",
        filename=os.path.join(HERE, "test_output", "user_supplied_wavelength"),
        stype="science",
        overwrite=True,
    )


peaks = np.sort(np.random.random(31) * 1000.0)
# Removed the closely spaced peaks
distance_mask = np.isclose(peaks[:-1], peaks[1:], atol=5.0)
distance_mask = np.insert(distance_mask, 0, False)
peaks = peaks[~distance_mask]

# Line list
wavelengths_linear = (
    3000.0 + 5.0 * peaks + (np.random.random(len(peaks)) - 0.5) * 2.0
)
wavelengths_quadratic = (
    3000.0
    + 4 * peaks
    + 1.0e-3 * peaks**2.0
    + (np.random.random(len(peaks)) - 0.5) * 2.0
)

elements_linear = ["Linear"] * len(wavelengths_linear)
elements_quadratic = ["Quadratic"] * len(wavelengths_quadratic)


def test_linear_fit():
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=1000,
        range_tolerance=500.0,
        xbins=200,
        ybins=200,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_linear, wavelengths=wavelengths_linear
    )
    wavecal.set_ransac_properties(minimum_matches=20)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(max_tries=500, fit_deg=1)
    # Refine solution
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.robust_refit(best_p, refine=False, robust_refit=True)

    assert np.abs(best_p[1] - 5.0) / 5.0 < 0.001
    assert np.abs(best_p[0] - 3000.0) / 3000.0 < 0.001
    assert peak_utilisation > 0.8
    assert atlas_utilisation > 0.0


def test_manual_refit():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=1000,
        range_tolerance=500.0,
        xbins=200,
        ybins=200,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_linear, wavelengths=wavelengths_linear
    )
    wavecal.set_ransac_properties(minimum_matches=25)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(max_tries=500, fit_deg=1)

    # Refine solution
    (
        best_p_robust,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.robust_refit(best_p, refine=False, robust_refit=True)

    (
        best_p_manual,
        matched_peaks,
        matched_atlas,
        rms,
        residuals,
    ) = wavecal.manual_refit(matched_peaks, matched_atlas)

    assert np.abs(best_p_manual[0] - best_p[0]) < 10.0
    assert np.abs(best_p_manual[1] - best_p[1]) < 0.1


def test_manual_refit_remove_points():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=1000,
        range_tolerance=500.0,
        xbins=200,
        ybins=200,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_linear, wavelengths=wavelengths_linear
    )
    wavecal.set_ransac_properties(minimum_matches=25)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        atched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(max_tries=500, fit_deg=1)

    # Refine solution
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.robust_refit(best_p, refine=False, robust_refit=True)

    wavecal.remove_pix_wave_pair(5)

    (
        best_p_manual,
        matched_peaks,
        matched_atlas,
        rms,
        residuals,
    ) = wavecal.manual_refit(matched_peaks, matched_atlas)

    assert np.allclose(best_p_manual, best_p)


def test_manual_refit_add_points():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=1000,
        range_tolerance=500.0,
        xbins=200,
        ybins=200,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_linear, wavelengths=wavelengths_linear
    )
    wavecal.set_ransac_properties(minimum_matches=25)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        atched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(max_tries=500, fit_deg=1)

    # Refine solution
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.robust_refit(best_p, refine=False, robust_refit=True)

    wavecal.add_pix_wave_pair(
        2000.0, 3000.0 + 4 * 2000.0 + 1.0e-3 * 2000.0**2.0
    )
    (
        best_p_manual,
        matched_peaks,
        matched_atlas,
        rms,
        residuals,
    ) = wavecal.manual_refit(matched_peaks, matched_atlas)

    assert np.allclose(best_p_manual, best_p)


def test_quadratic_fit():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=1000,
        range_tolerance=500.0,
        xbins=100,
        ybins=100,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_quadratic, wavelengths=wavelengths_quadratic
    )
    wavecal.set_ransac_properties(minimum_matches=20)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(
        max_tries=2000, fit_tolerance=5.0, candidate_tolerance=2.0, fit_deg=2
    )
    # Refine solution
    (
        best_p_robust,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.robust_refit(best_p, refine=False, robust_refit=True)


def test_quadratic_fit_legendre():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=500,
        range_tolerance=200.0,
        xbins=100,
        ybins=100,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_quadratic, wavelengths=wavelengths_quadratic
    )
    wavecal.set_ransac_properties(sample_size=10, minimum_matches=20)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(
        max_tries=2000,
        fit_tolerance=5.0,
        candidate_tolerance=2.0,
        fit_deg=2,
        fit_type="legendre",
    )


def test_quadratic_fit_chebyshev():
    # Initialise the calibrator
    wavecal = WavelengthCalibration(log_file_name=None)
    wavecal.initialise_calibrator(peaks)

    wavecal.set_calibrator_properties(num_pix=1000)
    wavecal.set_hough_properties(
        num_slopes=500,
        range_tolerance=200.0,
        xbins=100,
        ybins=100,
        min_wavelength=3000.0,
        max_wavelength=8000.0,
    )
    wavecal.add_user_atlas(
        elements=elements_quadratic, wavelengths=wavelengths_quadratic
    )
    wavecal.set_ransac_properties(sample_size=10, minimum_matches=20)
    wavecal.do_hough_transform(brute_force=False)

    # Run the wavelength calibration
    (
        best_p,
        matched_peaks,
        matched_atlas,
        rms,
        residual,
        peak_utilisation,
        atlas_utilisation,
    ) = wavecal.fit(
        max_tries=2000,
        fit_tolerance=5.0,
        candidate_tolerance=2.0,
        fit_deg=2,
        fit_type="chebyshev",
    )



# -*- coding: utf-8 -*-
import numpy as np
from aspired import spectrum_oneD
from rascal.calibrator import Calibrator


def test_spectrum_oned():
    spec = spectrum_oneD.SpectrumOneD(log_file_name=None)

    spec.add_count(count=np.arange(10))
    assert (spec.count == np.arange(10)).all()
    assert (spec.count_err == np.zeros(10)).all()
    assert (spec.count_sky == np.zeros(10)).all()
    spec.remove_count()
    assert spec.count is None
    assert spec.count_err is None
    assert spec.count_sky is None

    spec.add_variances(np.ones((100, 200)))
    assert (spec.var == np.ones((100, 200))).all()
    spec.remove_variances()
    assert spec.var is None

    spec.add_arc_spec(np.ones(100))
    assert (spec.arc_spec == np.ones(100)).all()
    spec.remove_arc_spec()
    assert spec.arc_spec is None

    spec.add_peaks(np.ones(100))
    assert (spec.peaks == np.ones(100)).all()
    spec.remove_peaks()
    assert spec.peaks is None

    spec.add_peaks_wave(np.ones(100))
    assert (spec.peaks_wave == np.ones(100)).all()
    spec.remove_peaks_wave()
    assert spec.peaks_wave is None

    spec.add_calibrator(Calibrator([100]))
    spec.remove_calibrator()
    assert spec.calibrator is None

    spec.add_atlas_wavelength_range(
        min_atlas_wavelength=1000.0, max_atlas_wavelength=10000.0
    )
    assert spec.min_atlas_wavelength == 1000.0
    assert spec.max_atlas_wavelength == 10000.0
    spec.remove_atlas_wavelength_range()
    assert spec.min_atlas_wavelength is None
    assert spec.max_atlas_wavelength is None

    spec.add_min_atlas_intensity(123456.0)
    assert spec.min_atlas_intensity == 123456.0
    spec.remove_min_atlas_intensity()
    assert spec.min_atlas_intensity is None

    spec.add_min_atlas_distance(0.123)
    assert spec.min_atlas_distance == 0.123
    spec.remove_min_atlas_distance()
    assert spec.min_atlas_distance is None

    spec.add_gain(2.71)
    assert spec.gain == 2.71
    spec.remove_gain()
    assert spec.gain is None

    spec.add_readnoise(10.234)
    assert spec.readnoise == 10.234
    spec.remove_readnoise()
    assert spec.readnoise is None

    spec.add_exptime(3600.0)
    assert spec.exptime == 3600.0
    spec.remove_exptime()
    assert spec.exptime is None

    spec.add_airmass(1.57)
    assert spec.airmass == 1.57
    spec.remove_airmass()
    assert spec.airmass is None

    spec.add_seeing(2.39)
    assert spec.seeing == 2.39
    spec.remove_seeing()
    assert spec.seeing is None

    spec.add_weather_condition(
        pressure=123456.0, temperature=279.3, relative_humidity=15.1
    )
    assert spec.pressure == 123456.0
    assert spec.temperature == 279.3
    assert spec.relative_humidity == 15.1
    spec.remove_weather_condition()
    assert spec.pressure is None
    assert spec.temperature is None
    assert spec.relative_humidity is None

    spec.add_fit_type("leg")
    assert spec.fit_type == "leg"
    spec.remove_fit_type()
    assert spec.fit_type is None

    spec.add_fit_coeff(np.arange(10))
    assert (spec.fit_coeff == np.arange(10)).all()
    spec.remove_fit_coeff()
    assert spec.fit_coeff is None

    spec.add_calibrator_properties(
        num_pix=1024,
        effective_pixel=np.arange(1024),
        plotting_library="plotly",
    )
    assert spec.num_pix == 1024
    assert (spec.effective_pixel == np.arange(1024)).all()
    assert spec.plotting_library == "plotly"
    assert spec.log_level == "info"
    spec.remove_calibrator_properties()
    assert spec.num_pix is None
    assert spec.effective_pixel is None
    assert spec.plotting_library is None

    spec.add_hough_properties(
        num_slopes=1000,
        xbins=120,
        ybins=250,
        min_wavelength=3000.0,
        max_wavelength=7800.0,
        range_tolerance=369.0,
        linearity_tolerance=135.0,
    )
    assert spec.num_slopes == 1000
    assert spec.xbins == 120
    assert spec.ybins == 250
    assert spec.min_wavelength == 3000.0
    assert spec.max_wavelength == 7800.0
    assert spec.range_tolerance == 369.0
    assert spec.linearity_tolerance == 135.0
    spec.remove_hough_properties()
    assert spec.num_slopes is None
    assert spec.xbins is None
    assert spec.ybins is None
    assert spec.min_wavelength is None
    assert spec.max_wavelength is None
    assert spec.range_tolerance is None
    assert spec.linearity_tolerance is None

    spec.add_ransac_properties(
        sample_size=999,
        top_n_candidate=7,
        linear=True,
        filter_close=True,
        ransac_tolerance=5.0,
        candidate_weighted=True,
        hough_weight=1.3,
        minimum_matches=5,
        minimum_peak_utilisation=80.0,
        minimum_fit_error=0.1,
    )
    assert spec.sample_size == 999
    assert spec.top_n_candidate == 7
    assert spec.linear
    assert spec.filter_close
    assert spec.ransac_tolerance == 5.0
    assert spec.candidate_weighted
    assert spec.hough_weight == 1.3
    assert spec.minimum_matches == 5
    assert spec.minimum_peak_utilisation == 80.0
    assert spec.minimum_fit_error == 0.1
    spec.remove_ransac_properties()
    assert spec.sample_size is None
    assert spec.top_n_candidate is None
    assert spec.linear is None
    assert spec.filter_close is None
    assert spec.ransac_tolerance is None
    assert spec.candidate_weighted is None
    assert spec.hough_weight is None
    assert spec.minimum_matches is None
    assert spec.minimum_peak_utilisation is None
    assert spec.minimum_fit_error is None

    spec.add_fit_output_final(
        fit_coeff=[1, 2, 5, 7, 10],
        matched_peaks=[0, 1, 2, 3],
        matched_atlas=[10, 11, 12, 13],
        rms=0.123456,
        residual=0.56789,
        peak_utilisation=87.67894,
        atlas_utilisation=51.7643,
    )
    assert spec.fit_coeff == [1, 2, 5, 7, 10]
    assert spec.matched_peaks == [0, 1, 2, 3]
    assert spec.matched_atlas == [10, 11, 12, 13]
    assert spec.rms == 0.123456
    assert spec.residual == 0.56789
    assert spec.peak_utilisation == 87.67894
    assert spec.atlas_utilisation == 51.7643
    spec.remove_fit_output_final()
    assert spec.fit_coeff is None
    assert spec.matched_peaks is None
    assert spec.matched_atlas is None
    assert spec.rms is None
    assert spec.residual is None
    assert spec.peak_utilisation is None
    assert spec.atlas_utilisation is None

    spec.add_fit_output_rascal(
        fit_coeff=[1, 2, 5, 7, 10],
        matched_peaks=[0, 1, 2, 3],
        matched_atlas=[10, 11, 12, 13],
        rms=0.123456,
        residual=0.56789,
        peak_utilisation=87.67894,
        atlas_utilisation=51.7643,
    )
    assert spec.fit_coeff_rascal == [1, 2, 5, 7, 10]
    assert spec.matched_peaks == [0, 1, 2, 3]
    assert spec.matched_atlas == [10, 11, 12, 13]
    assert spec.rms_rascal == 0.123456
    assert spec.residual_rascal == 0.56789
    assert spec.peak_utilisation_rascal == 87.67894
    assert spec.atlas_utilisation_rascal == 51.7643
    spec.remove_fit_output_rascal()
    assert spec.fit_coeff_rascal is None
    assert spec.matched_peaks_rascal is None
    assert spec.matched_atlas_rascal is None
    assert spec.rms_rascal is None
    assert spec.residual_rascal is None
    assert spec.peak_utilisation_rascal is None
    assert spec.atlas_utilisation_rascal is None

    spec.add_fit_output_refine(
        fit_coeff=[1, 2, 5, 7, 10],
        matched_peaks=[0, 1, 2, 3],
        matched_atlas=[10, 11, 12, 13],
        rms=0.123456,
        residual=0.56789,
        peak_utilisation=87.67894,
        atlas_utilisation=51.7643,
    )
    assert spec.fit_coeff_refine == [1, 2, 5, 7, 10]
    assert spec.matched_peaks_refine == [0, 1, 2, 3]
    assert spec.matched_atlas_refine == [10, 11, 12, 13]
    assert spec.rms_refine == 0.123456
    assert spec.residual_refine == 0.56789
    assert spec.peak_utilisation_refine == 87.67894
    assert spec.atlas_utilisation_refine == 51.7643
    spec.remove_fit_output_refine()
    assert spec.fit_coeff_refine is None
    assert spec.matched_peaks_refine is None
    assert spec.matched_atlas_refine is None
    assert spec.rms_refine is None
    assert spec.residual_refine is None
    assert spec.peak_utilisation_refine is None
    assert spec.atlas_utilisation_refine is None

    spec.add_wavelength(np.arange(1000))
    assert (spec.wave == np.arange(1000)).all()
    spec.remove_wavelength()
    assert spec.wave is None

    spec.add_wavelength_resampled(np.arange(789))
    assert spec.wave_bin == 1
    assert spec.wave_start == 0
    assert spec.wave_end == 788
    assert (spec.wave_resampled == np.arange(789)).all()
    spec.remove_wavelength_resampled()
    assert spec.wave_bin is None
    assert spec.wave_start is None
    assert spec.wave_end is None
    assert spec.wave_resampled is None

    spec.add_count_resampled(np.arange(100), np.arange(200), np.arange(300))
    assert (spec.count_resampled == np.arange(100)).all()
    assert (spec.count_err_resampled == np.arange(200)).all()
    assert (spec.count_sky_resampled == np.arange(300)).all()
    spec.remove_count_resampled()
    assert spec.count_resampled is None
    assert spec.count_err_resampled is None
    assert spec.count_sky_resampled is None

    spec.add_smoothing(smooth=True)
    assert spec.smooth
    spec.remove_smoothing()
    assert spec.smooth is None

    spec.add_sensitivity_func(np.poly1d)
    assert spec.sensitivity_func == np.poly1d
    spec.remove_sensitivity_func()
    assert spec.sensitivity_func is None

    spec.add_sensitivity(np.arange(1000))
    assert (spec.sensitivity == np.arange(1000)).all()
    spec.remove_sensitivity()
    assert spec.sensitivity is None

    spec.add_sensitivity_resampled(np.arange(2000))
    assert (spec.sensitivity_resampled == np.arange(2000)).all()
    spec.remove_sensitivity_resampled()
    assert spec.sensitivity_resampled is None

    spec.add_literature_standard(
        wave_literature=np.arange(456), flux_literature=np.ones(789)
    )
    assert (spec.wave_literature == np.arange(456)).all()
    assert (spec.flux_literature == np.ones(789)).all()
    spec.remove_literature_standard()
    assert spec.wave_literature is None
    assert spec.flux_literature is None

    spec.add_flux(np.arange(100), np.arange(200), np.arange(300))
    assert (spec.flux == np.arange(100)).all()
    assert (spec.flux_err == np.arange(200)).all()
    assert (spec.flux_sky == np.arange(300)).all()
    spec.remove_flux()
    assert spec.flux is None
    assert spec.flux_err is None
    assert spec.flux_sky is None

    spec.add_atm_ext(np.arange(100))
    assert (spec.atm_ext == np.arange(100)).all()
    spec.remove_atm_ext()
    assert spec.atm_ext is None

    spec.add_flux_atm_ext_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (spec.flux_atm_ext_corrected == np.arange(100)).all()
    assert (spec.flux_err_atm_ext_corrected == np.arange(200)).all()
    assert (spec.flux_sky_atm_ext_corrected == np.arange(300)).all()
    spec.remove_flux_atm_ext_corrected()
    assert spec.flux_atm_ext_corrected is None
    assert spec.flux_err_atm_ext_corrected is None
    assert spec.flux_sky_atm_ext_corrected is None

    spec.add_telluric_profile(np.arange(100))
    assert (spec.telluric_profile == np.arange(100)).all()
    spec.remove_telluric_profile()
    assert spec.telluric_profile is None

    spec.add_flux_telluric_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (spec.flux_telluric_corrected == np.arange(100)).all()
    assert (spec.flux_err_telluric_corrected == np.arange(200)).all()
    assert (spec.flux_sky_telluric_corrected == np.arange(300)).all()
    spec.remove_flux_telluric_corrected()
    assert spec.flux_telluric_corrected is None
    assert spec.flux_err_telluric_corrected is None
    assert spec.flux_sky_telluric_corrected is None

    spec.add_flux_atm_ext_telluric_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (spec.flux_atm_ext_telluric_corrected == np.arange(100)).all()
    assert (spec.flux_err_atm_ext_telluric_corrected == np.arange(200)).all()
    assert (spec.flux_sky_atm_ext_telluric_corrected == np.arange(300)).all()
    spec.remove_flux_atm_ext_telluric_corrected()
    assert spec.flux_atm_ext_telluric_corrected is None
    assert spec.flux_err_atm_ext_telluric_corrected is None
    assert spec.flux_sky_atm_ext_telluric_corrected is None

    spec.add_flux_resampled(np.arange(100), np.arange(200), np.arange(300))
    assert (spec.flux_resampled == np.arange(100)).all()
    assert (spec.flux_err_resampled == np.arange(200)).all()
    assert (spec.flux_sky_resampled == np.arange(300)).all()
    spec.remove_flux_resampled()
    assert spec.flux_resampled is None
    assert spec.flux_err_resampled is None
    assert spec.flux_sky_resampled is None

    spec.add_atm_ext_resampled(np.arange(100))
    assert (spec.atm_ext_resampled == np.arange(100)).all()
    spec.remove_atm_ext_resampled()
    assert spec.atm_ext_resampled is None

    spec.add_flux_resampled_atm_ext_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (spec.flux_resampled_atm_ext_corrected == np.arange(100)).all()
    assert (spec.flux_err_resampled_atm_ext_corrected == np.arange(200)).all()
    assert (spec.flux_sky_resampled_atm_ext_corrected == np.arange(300)).all()
    spec.remove_flux_resampled_atm_ext_corrected()
    assert spec.flux_resampled_atm_ext_corrected is None
    assert spec.flux_err_resampled_atm_ext_corrected is None
    assert spec.flux_sky_resampled_atm_ext_corrected is None

    spec.add_telluric_profile_resampled(np.arange(100))
    assert (spec.telluric_profile_resampled == np.arange(100)).all()
    spec.remove_telluric_profile_resampled()
    assert spec.telluric_profile_resampled is None

    spec.add_flux_resampled_telluric_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (spec.flux_resampled_telluric_corrected == np.arange(100)).all()
    assert (spec.flux_err_resampled_telluric_corrected == np.arange(200)).all()
    assert (spec.flux_sky_resampled_telluric_corrected == np.arange(300)).all()
    spec.remove_flux_resampled_telluric_corrected()
    assert spec.flux_resampled_telluric_corrected is None
    assert spec.flux_err_resampled_telluric_corrected is None
    assert spec.flux_sky_resampled_telluric_corrected is None

    spec.add_flux_resampled_atm_ext_telluric_corrected(
        np.arange(100), np.arange(200), np.arange(300)
    )
    assert (
        spec.flux_resampled_atm_ext_telluric_corrected == np.arange(100)
    ).all()
    assert (
        spec.flux_err_resampled_atm_ext_telluric_corrected == np.arange(200)
    ).all()
    assert (
        spec.flux_sky_resampled_atm_ext_telluric_corrected == np.arange(300)
    ).all()
    spec.remove_flux_resampled_atm_ext_telluric_corrected()
    assert spec.flux_resampled_atm_ext_telluric_corrected is None
    assert spec.flux_err_resampled_atm_ext_telluric_corrected is None
    assert spec.flux_sky_resampled_atm_ext_telluric_corrected is None



# -*- coding: utf-8 -*-
import os
from unittest.mock import patch

import numpy as np
import pytest
from aspired import image_reduction
from astropy.io import fits

HERE = os.path.dirname(os.path.realpath(__file__))


def file_len(fname):
    with open(fname) as f:
        for i, _ in enumerate(f):
            pass
    return i + 1


def test_logger():
    imred_debug = image_reduction.ImageReducer(
        log_level="DEBUG",
        logger_name="imred_debug",
        log_file_name="imred_debug.log",
        log_file_folder=os.path.join(HERE, "test_output"),
    )
    imred_info = image_reduction.ImageReducer(
        log_level="INFO",
        logger_name="imred_info",
        log_file_name="imred_info.log",
        log_file_folder=os.path.join(HERE, "test_output"),
    )
    imred_warning = image_reduction.ImageReducer(
        log_level="WARNING",
        logger_name="imred_warning",
        log_file_name="imred_warning.log",
        log_file_folder=os.path.join(HERE, "test_output"),
    )
    imred_error = image_reduction.ImageReducer(
        log_level="ERROR",
        logger_name="imred_error",
        log_file_name="imred_error.log",
        log_file_folder=os.path.join(HERE, "test_output"),
    )
    imred_critical = image_reduction.ImageReducer(
        log_level="CRITICAL",
        logger_name="imred_critical",
        log_file_name="imred_critical.log",
        log_file_folder=os.path.join(HERE, "test_output"),
    )

    imred_debug.logger.debug("debug: debug mode")
    imred_debug.logger.info("debug: info mode")
    imred_debug.logger.warning("debug: warning mode")
    imred_debug.logger.error("debug: error mode")
    imred_debug.logger.critical("debug: critical mode")

    imred_info.logger.debug("info: debug mode")
    imred_info.logger.info("info: info mode")
    imred_info.logger.warning("info: warning mode")
    imred_info.logger.error("info: error mode")
    imred_info.logger.critical("info: critical mode")

    imred_warning.logger.debug("warning: debug mode")
    imred_warning.logger.info("warning: info mode")
    imred_warning.logger.warning("warning: warning mode")
    imred_warning.logger.error("warning: error mode")
    imred_warning.logger.critical("warning: critical mode")

    imred_error.logger.debug("error: debug mode")
    imred_error.logger.info("error: info mode")
    imred_error.logger.warning("error: warning mode")
    imred_error.logger.error("error: error mode")
    imred_error.logger.critical("error: critical mode")

    imred_critical.logger.debug("critical: debug mode")
    imred_critical.logger.info("critical: info mode")
    imred_critical.logger.warning("critical: warning mode")
    imred_critical.logger.error("critical: error mode")
    imred_critical.logger.critical("critical: critical mode")

    debug_debug_length = file_len(
        os.path.join(HERE, "test_output", "imred_debug.log")
    )
    debug_info_length = file_len(
        os.path.join(HERE, "test_output", "imred_info.log")
    )
    debug_warning_length = file_len(
        os.path.join(HERE, "test_output", "imred_warning.log")
    )
    debug_error_length = file_len(
        os.path.join(HERE, "test_output", "imred_error.log")
    )
    debug_critical_length = file_len(
        os.path.join(HERE, "test_output", "imred_critical.log")
    )

    assert (
        debug_debug_length == 5
    ), "Expecting 5 lines in the log file, " + "{} is logged.".format(
        debug_debug_length
    )
    assert (
        debug_info_length == 4
    ), "Expecting 4 lines in the log file, " + "{} is logged.".format(
        debug_info_length
    )
    assert (
        debug_warning_length == 3
    ), "Expecting 3 lines in the log file, " + "{} is logged.".format(
        debug_warning_length
    )
    assert (
        debug_error_length == 2
    ), "Expecting 2 lines in the log file, " + "{} is logged.".format(
        debug_error_length
    )
    assert (
        debug_critical_length == 1
    ), "Expecting 1 lines in the log file, " + "{} is logged.".format(
        debug_critical_length
    )


try:
    os.remove("test/test_output/imred_debug.log")
except Exception as e:
    print(e)

try:
    os.remove("test/test_output/imred_info.log")
except Exception as e:
    print(e)

try:
    os.remove("test/test_output/imred_warning.log")
except Exception as e:
    print(e)

try:
    os.remove("test/test_output/imred_error.log")
except Exception as e:
    print(e)

try:
    os.remove("test/test_output/imred_critical.log")
except Exception as e:
    print(e)


@pytest.mark.xfail()
def test_str_verbose_at_initialisation():
    image_reduction.ImageReducer(verbose="haha", log_file_name=None)


def test_absolute_path():
    current_absolute_path = os.path.abspath(
        "test/test_data/sprat_LHS6328.list"
    )
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(filelist=current_absolute_path)
    img.load_data()
    img.reduce()


def test_absolute_path_with_deprecation_warning_for_ImageReduction():
    current_absolute_path = os.path.abspath(
        "test/test_data/sprat_LHS6328.list"
    )
    img = image_reduction.ImageReduction(log_file_name=None)
    img.add_filelist(filelist=current_absolute_path)
    img.load_data()
    img.reduce()


def test_space_separated_input():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.txt"),
        ftype="ascii",
    )
    img.load_data()
    img.reduce()


def test_space_separated_input_2():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.txt"),
        delimiter=" ",
    )
    img.load_data()
    img.reduce()


def test_tsv_input_2():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.tsv"),
        ftype="tsv",
        delimiter="\t",
    )
    img.load_data()
    img.reduce()


def test_input_with_hdu():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328_with_hdu.list")
    )
    img.load_data()
    img.reduce()


def test_input_with_extra_bracket():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE, "test_data", "sprat_LHS6328_fake_extra_bracket.list"
        )
    )
    img.load_data()
    img.reduce()


def test_input_with_data_cube():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE, "test_data", "sprat_LHS6328_fake_data_cube.list"
        )
    )
    img.load_data()
    img.reduce()


def test_input_with_data_cube_extra_bracket():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE,
            "test_data",
            "sprat_LHS6328_fake_data_cube_extra_bracket.list",
        )
    )
    img.load_data()
    img.reduce()


@pytest.mark.xfail()
def test_input_with_one_dimensional_data():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE, "test_data", "sprat_LHS6328_one_dimensional_data.list"
        )
    )
    img.load_data()


def test_input_with_wrong_light_combine_type():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.load_data()
    img.set_properties(combinetype_light="FAIL")


def test_input_with_wrong_dark_combine_type():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(combinetype_dark="FAIL")
    img.load_data()
    img.reduce()


def test_set_everything_to_None():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        saxis=None,
        combinetype_light=None,
        sigma_clipping_light=None,
        clip_low_light=None,
        clip_high_light=None,
        exptime_light=None,
        exptime_light_keyword=None,
        combinetype_arc=None,
        sigma_clipping_arc=None,
        clip_low_arc=None,
        clip_high_arc=None,
        combinetype_dark=None,
        sigma_clipping_dark=None,
        clip_low_dark=None,
        clip_high_dark=None,
        exptime_dark=None,
        exptime_dark_keyword=None,
        combinetype_bias=None,
        sigma_clipping_bias=None,
        clip_low_bias=None,
        clip_high_bias=None,
        combinetype_flat=None,
        sigma_clipping_flat=None,
        clip_low_flat=None,
        clip_high_flat=None,
        exptime_flat=None,
        exptime_flat_keyword=None,
        cosmicray=None,
        gain=None,
        readnoise=None,
        fsmode=None,
        psfmodel=None,
        cutoff=None,
        grow=None,
        iterations=None,
        diagonal=None,
    )
    img.load_data()
    img.reduce()


def test_set_everything_to_minus1():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        saxis=-1,
        combinetype_light=-1,
        sigma_clipping_light=-1,
        clip_low_light=-1,
        clip_high_light=-1,
        exptime_light=-1,
        exptime_light_keyword=-1,
        combinetype_arc=-1,
        sigma_clipping_arc=-1,
        clip_low_arc=-1,
        clip_high_arc=-1,
        combinetype_dark=-1,
        sigma_clipping_dark=-1,
        clip_low_dark=-1,
        clip_high_dark=-1,
        exptime_dark=-1,
        exptime_dark_keyword=-1,
        combinetype_bias=-1,
        sigma_clipping_bias=-1,
        clip_low_bias=-1,
        clip_high_bias=-1,
        combinetype_flat=-1,
        sigma_clipping_flat=-1,
        clip_low_flat=-1,
        clip_high_flat=-1,
        exptime_flat=-1,
        exptime_flat_keyword=-1,
        cosmicray=-1,
        gain=-1,
        readnoise=-1,
        fsmode=-1,
        psfmodel=-1,
        cutoff=-1,
        grow=-1,
        iterations=-1,
        diagonal=-1,
    )
    img.load_data()
    img.reduce()


# apart from heal_pixels
def test_set_everything_to_True():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        saxis=True,
        combinetype_light=True,
        sigma_clipping_light=True,
        clip_low_light=True,
        clip_high_light=True,
        exptime_light=True,
        exptime_light_keyword=True,
        combinetype_arc=True,
        sigma_clipping_arc=True,
        clip_low_arc=True,
        clip_high_arc=True,
        combinetype_dark=True,
        sigma_clipping_dark=True,
        clip_low_dark=True,
        clip_high_dark=True,
        exptime_dark=True,
        exptime_dark_keyword=True,
        combinetype_bias=True,
        sigma_clipping_bias=True,
        clip_low_bias=True,
        clip_high_bias=True,
        combinetype_flat=True,
        sigma_clipping_flat=True,
        clip_low_flat=True,
        clip_high_flat=True,
        exptime_flat=True,
        exptime_flat_keyword=True,
        cosmicray=True,
        gain=True,
        readnoise=True,
        fsmode=True,
        psfmodel=True,
        heal_pixels=False,
        cutoff=True,
        grow=True,
        iterations=True,
        diagonal=True,
    )
    img.load_data()
    img.reduce()


def test_set_everything_to_bla():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        saxis="bla",
        combinetype_light="bla",
        sigma_clipping_light="bla",
        clip_low_light="bla",
        clip_high_light="bla",
        exptime_light="bla",
        exptime_light_keyword="bla",
        combinetype_arc="bla",
        sigma_clipping_arc="bla",
        clip_low_arc="bla",
        clip_high_arc="bla",
        combinetype_dark="bla",
        sigma_clipping_dark="bla",
        clip_low_dark="bla",
        clip_high_dark="bla",
        exptime_dark="bla",
        exptime_dark_keyword="bla",
        combinetype_bias="bla",
        sigma_clipping_bias="bla",
        clip_low_bias="bla",
        clip_high_bias="bla",
        combinetype_flat="bla",
        sigma_clipping_flat="bla",
        clip_low_flat="bla",
        clip_high_flat="bla",
        exptime_flat="bla",
        exptime_flat_keyword="bla",
        cosmicray="bla",
        gain="bla",
        readnoise="bla",
        fsmode="bla",
        psfmodel="bla",
        heal_pixels="bla",
        cutoff="bla",
        grow="bla",
        iterations="bla",
        diagonal="bla",
    )
    img.load_data()
    img.reduce()


# The bad combinetype does not affect the reduction because
# there is not any flat frame
def test_input_with_wrong_flat_combine_type():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(combinetype_flat="FAIL")
    img.load_data()
    img.reduce()


def test_input_with_wrong_bias_combine_type():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(combinetype_bias="FAIL")
    img.load_data()
    img.reduce()


@pytest.mark.xfail(raises=RuntimeError)
def test_input_with_only_one_column():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE, "test_data", "sprat_LHS6328_expect_fail.list"
        )
    )


def test_cosmicray_cleaning_x_then_y():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        cosmicray=True, psfmodel="gaussxy", combinetype_light="average"
    )
    img.load_data()
    img.reduce()


def test_cosmicray_cleaning_y_then_x():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        cosmicray=True, psfmodel="gaussyx", combinetype_dark="average"
    )
    img.load_data()
    img.reduce()


def test_cosmicray_cleaning():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.set_properties(
        cosmicray=True, psfmodel="gaussx", combinetype_bias="average"
    )
    img.load_data()
    img.reduce()


def test_input_with_multiple_frames_to_combine_and_reduction():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(
            HERE, "test_data", "sprat_LHS6328_repeated_data.list"
        )
    )
    img.set_properties(combinetype_flat="average")
    img.load_data()
    img.reduce()


def test_input_with_one_line_and_reduction():
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328_one_line.list")
    )
    img.load_data()
    img.reduce()


def test_input_with_numpy_array_and_reduction():
    filelist = np.loadtxt(
        os.path.join(HERE, "test_data", "sprat_LHS6328.list"),
        delimiter=",",
        dtype="object",
    )
    for i, filepath in enumerate(filelist[:, 1]):
        filelist[:, 1][i] = os.path.join(HERE, "test_data", filepath.strip())
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(filelist=filelist)
    img.load_data()
    img.reduce()


@patch("plotly.graph_objects.Figure.show")
def test_reduction_and_save(mock_show):
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(
        filelist=os.path.join(HERE, "test_data", "sprat_LHS6328.list")
    )
    img.load_data()
    img.reduce()
    img.save_fits(
        os.path.join(HERE, "test_output", "reduced_image"), overwrite=True
    )
    img.inspect(
        display=False,
        filename=os.path.join(HERE, "test_output", "reduced_image"),
        save_fig=True,
        fig_type="iframe+png+svg+pdf+jpg",
    )
    img.inspect(display=True)
    img.list_files()


def test_input_with_numpy_array_and_clean_bad_pixels():
    filelist = np.loadtxt(
        os.path.join(HERE, "test_data", "sprat_LHS6328.list"),
        delimiter=",",
        dtype="object",
    )
    for i, filepath in enumerate(filelist[:, 1]):
        filelist[:, 1][i] = os.path.join(
            os.path.join(HERE, "test_data"), filepath.strip()
        )
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(filelist=filelist)
    img.load_data()
    img.reduce()
    img.heal_bad_pixels()


def test_input_with_numpy_array_and_set_every_pixel_bad():
    filelist = np.loadtxt(
        os.path.join(HERE, "test_data", "sprat_LHS6328.list"),
        delimiter=",",
        dtype="object",
    )
    for i, filepath in enumerate(filelist[:, 1]):
        filelist[:, 1][i] = os.path.join(
            os.path.join(HERE, "test_data"), filepath.strip()
        )
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(filelist=filelist)
    img.load_data()
    img.reduce()
    img.heal_bad_pixels(np.zeros_like(img.light_reduced))


@pytest.mark.xfail(raises=RuntimeError)
def test_input_with_numpy_array_and_clean_bad_pixels_expect_fail():
    filelist = np.loadtxt(
        os.path.join(HERE, "test_data", "sprat_LHS6328.list"),
        delimiter=",",
        dtype="object",
    )
    for i, filepath in enumerate(filelist[:, 1]):
        filelist[:, 1][i] = os.path.join(
            os.path.join(HERE, "test_data"), filepath.strip()
        )
    img = image_reduction.ImageReducer(log_file_name=None)
    img.add_filelist(filelist=filelist)
    img.load_data()
    img.reduce()
    img.create_bad_mask()
    img.heal_bad_pixels(bad_mask=np.ones(100))


def test_input_with_fits_data_object():
    filelist = np.loadtxt(
        os.path.join(HERE, "test_data", "sprat_LHS6328.list"),
        delimiter=",",
        dtype="object",
    )
    for i, filepath in enumerate(filelist[:, 1]):
        filelist[:, 1][i] = os.path.join(
            os.path.join(HERE, "test_data"), filepath.strip()
        )
    img = image_reduction.ImageReducer(log_file_name=None)
    imtype = filelist[:, 0].astype("object")
    impath = filelist[:, 1].astype("object")

    dark_list = impath[imtype == "dark"]
    arc_list = impath[imtype == "arc"]
    light_list = impath[imtype == "light"]

    for i in range(light_list.size):
        light = fits.open(light_list[i])[0]
        img.add_light(light.data, light.header, float(light.header["EXPTIME"]))

    for i in range(arc_list.size):
        arc = fits.open(arc_list[i])[0]
        img.add_arc(arc.data, arc.header)

    for i in range(dark_list.size):
        dark = fits.open(dark_list[i])[0]
        img.add_dark(dark.data, dark.header, float(dark.header["EXPTIME"]))

    img.reduce()



# -*- coding: utf-8 -*-
import os
from unittest.mock import patch

import numpy as np
from aspired import image_reduction, spectral_reduction

HERE = os.path.dirname(os.path.realpath(__file__))


@patch("plotly.graph_objects.Figure.show")
def test_full_run(mock_show):
    # Extract two spectra

    # Line list
    atlas = [
        4193.5,
        4385.77,
        4500.98,
        4524.68,
        4582.75,
        4624.28,
        4671.23,
        4697.02,
        4734.15,
        4807.02,
        4921.48,
        5028.28,
        5618.88,
        5823.89,
        5893.29,
        5934.17,
        6182.42,
        6318.06,
        6472.841,
        6595.56,
        6668.92,
        6728.01,
        6827.32,
        6976.18,
        7119.60,
        7257.9,
        7393.8,
        7584.68,
        7642.02,
        7740.31,
        7802.65,
        7887.40,
        7967.34,
        8057.258,
    ]
    element = ["Xe"] * len(atlas)

    spatial_mask = np.arange(20, 200)
    spec_mask = np.arange(100, 1024)

    # Science frame
    lhs6328_frame = image_reduction.ImageReduction(
        log_level="ERROR", log_file_folder=os.path.join(HERE, "test_output")
    )
    lhs6328_frame.add_filelist("test/test_data/sprat_LHS6328.list")
    lhs6328_frame.load_data()
    lhs6328_frame.reduce()
    lhs6328_frame.save_fits(
        "test/test_output/test_full_run_standard_image", overwrite=True
    )

    lhs6328_twodspec = spectral_reduction.TwoDSpec(
        lhs6328_frame,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        readnoise=5.7,
        log_level="ERROR",
        log_file_folder=os.path.join(HERE, "test_output"),
    )

    lhs6328_twodspec.ap_trace(nspec=2, display=False)

    # Optimal extraction to get the LSF for force extraction below
    lhs6328_twodspec.ap_extract(
        apwidth=15, skywidth=10, skydeg=1, optimal=True, display=False
    )
    lhs6328_twodspec.inspect_line_spread_function(
        display=False,
        filename=os.path.join(
            HERE, "test_output", "test_full_run_line_spread_function_gauss"
        ),
        save_fig=True,
        fig_type="iframe+png",
        return_jsonstring=True,
    )

    # Force extraction
    lhs6328_twodspec.ap_extract(
        apwidth=15,
        skywidth=10,
        skydeg=1,
        optimal=True,
        forced=True,
        variances=lhs6328_twodspec.spectrum_list[1].var,
        display=False,
    )

    # Aperture extraction
    lhs6328_twodspec.ap_extract(
        apwidth=15,
        skywidth=5,
        skydeg=1,
        optimal=False,
        display=False,
        save_fig=True,
        filename=os.path.join(HERE, "test_output", "test_full_run_extract"),
    )
    lhs6328_twodspec.inspect_line_spread_function(
        spec_id=0, display=False, return_jsonstring=True
    )
    lhs6328_twodspec.inspect_line_spread_function(
        display=False, return_jsonstring=True
    )

    # Optimal extraction
    lhs6328_twodspec.ap_extract(
        apwidth=15,
        skywidth=5,
        skydeg=1,
        optimal=True,
        forced=True,
        variances=10000.0,
        display=False,
        save_fig=True,
        filename=os.path.join(HERE, "test_output", "test_full_run_extract"),
    )

    lhs6328_twodspec.inspect_line_spread_function(
        save_fig=True,
        display=True,
        open_iframe=False,
        filename=os.path.join(HERE, "test_output", "test_full_run_lsf"),
    )

    lhs6328_twodspec.save_fits(
        filename=os.path.join(HERE, "test_output", "test_full_run_twospec"),
        overwrite=True,
    )

    # Standard frame
    standard_frame = image_reduction.ImageReduction(
        log_level="ERROR", log_file_folder=os.path.join(HERE, "test_output")
    )
    standard_frame.add_filelist(
        os.path.join(HERE, "test_data", "sprat_Hiltner102.list")
    )
    standard_frame.load_data()
    standard_frame.reduce()
    standard_frame.save_fits(
        os.path.join(HERE, "test_output", "test_full_run_standard_image"),
        overwrite=True,
    )

    hilt102_twodspec = spectral_reduction.TwoDSpec(
        standard_frame,
        spatial_mask=spatial_mask,
        spec_mask=spec_mask,
        readnoise=5.7,
        log_level="ERROR",
        log_file_folder=os.path.join(HERE, "test_output"),
    )

    hilt102_twodspec.ap_trace(
        nspec=1,
        resample_factor=10,
        display=False,
        save_fig=True,
        fig_type="jpg+png+pdf+svg+iframe",
        filename=os.path.join(HERE, "test_output", "test_full_run_aptrace"),
    )

    hilt102_twodspec.ap_extract(
        apwidth=15, skysep=3, skywidth=5, skydeg=1, optimal=True, display=False
    )
    hilt102_twodspec.inspect_line_spread_function(
        display=True,
        save_fig=True,
        fig_type="jpg+png+pdf+svg+iframe",
        filename=os.path.join(HERE, "test_output", "test_full_run_lsf"),
        return_jsonstring=True,
    )

    # Extract the 1D arc by aperture sum of the traces provided
    lhs6328_twodspec.extract_arc_spec(
        display=True,
        save_fig=True,
        fig_type="jpg+png+pdf+svg+iframe",
        filename=os.path.join(HERE, "test_output", "test_full_run_arc_spec"),
        return_jsonstring=True,
    )
    hilt102_twodspec.extract_arc_spec(display=True)
    hilt102_twodspec.create_fits(output="trace+count")

    # Handle 1D Science spectrum
    lhs6328_onedspec = spectral_reduction.OneDSpec(
        log_level="ERROR", log_file_folder=os.path.join(HERE, "test_output")
    )
    lhs6328_onedspec.from_twodspec(lhs6328_twodspec, stype="science")
    lhs6328_onedspec.from_twodspec(hilt102_twodspec, stype="standard")

    # Create the trace and count as FITS BEFORE flux and wavelength calibration
    # This is an uncommon operation, but it should work.
    lhs6328_onedspec.create_fits(
        output="trace+count", stype="science+standard"
    )

    # Save the trace and count as FITS BEFORE flux and wavelength calibration
    # This is an uncommon operation, but it should work.
    lhs6328_onedspec.save_fits(
        output="trace+count",
        filename=os.path.join(HERE, "test_output", "test_full_run"),
        stype="science+standard",
        overwrite=True,
    )

    # Find the peaks of the arc
    lhs6328_onedspec.find_arc_lines(display=True, stype="science+standard")
    lhs6328_onedspec.find_arc_lines(display=False, stype="science+standard")
    lhs6328_onedspec.inspect_arc_lines(
        spec_id=0, display=True, stype="science+standard"
    )
    lhs6328_onedspec.inspect_arc_lines(
        spec_id=0, display=False, stype="science+standard"
    )

    # Configure the wavelength calibrator
    lhs6328_onedspec.initialise_calibrator(stype="science+standard")

    lhs6328_onedspec.set_hough_properties(
        num_slopes=2000,
        xbins=200,
        ybins=200,
        min_wavelength=3700,
        max_wavelength=8200,
        stype="science+standard",
    )
    lhs6328_onedspec.set_ransac_properties(
        filter_close=True, stype="science+standard"
    )

    lhs6328_onedspec.add_user_atlas(
        elements=element, wavelengths=atlas, stype="science+standard"
    )
    lhs6328_onedspec.do_hough_transform()

    # Solve for the pixel-to-wavelength solution
    lhs6328_onedspec.fit(
        max_tries=200, stype="science+standard", display=False
    )
    lhs6328_onedspec.robust_refit()

    # list the matched pixel-peaks
    lhs6328_onedspec.get_pix_wave_pairs(spec_id=0)
    lhs6328_onedspec.get_pix_wave_pairs()

    # get all the calibrators
    _ = lhs6328_onedspec.get_calibrator()

    # Apply the wavelength calibration and display it
    lhs6328_onedspec.apply_wavelength_calibration(stype="science+standard")

    # Get the standard from the library
    lhs6328_onedspec.load_standard(target="hiltner102")

    lhs6328_onedspec.get_sensitivity(
        sens_deg=11, method="polynomial", mask_fit_size=1
    )

    lhs6328_onedspec.get_sensitivity(
        k=3, method="interpolate", mask_fit_size=1
    )

    lhs6328_onedspec.apply_flux_calibration(stype="science+standard")
    lhs6328_onedspec.inspect_reduced_spectrum()

    lhs6328_onedspec.get_flux_continuum()

    lhs6328_onedspec.get_telluric_profile()
    lhs6328_onedspec.inspect_telluric_profile(display=False)
    lhs6328_onedspec.apply_telluric_correction()
    lhs6328_onedspec.inspect_reduced_spectrum()

    # Apply atmospheric extinction correction
    lhs6328_onedspec.set_atmospheric_extinction(location="orm")
    lhs6328_onedspec.apply_atmospheric_extinction_correction()
    lhs6328_onedspec.inspect_reduced_spectrum()

    # Create FITS
    lhs6328_onedspec.create_fits(output="trace+count")

    # Modify FITS header for the trace
    lhs6328_onedspec.modify_trace_header(0, "set", "COMMENT", "Hello World!")

    print(
        lhs6328_onedspec.science_spectrum_list[0]
        .trace_hdulist[0]
        .header["COMMENT"]
    )

    # Create more FITS
    lhs6328_onedspec.create_fits(
        output="trace+count+arc_spec+wavecal_coefficients",
        stype="science+standard",
        recreate=False,
    )

    # Check the modified headers are not overwritten
    print(
        lhs6328_onedspec.science_spectrum_list[0]
        .trace_hdulist[0]
        .header["COMMENT"]
    )

    # Save as FITS (and create the ones that were not created earlier)
    lhs6328_onedspec.save_fits(
        output="*",
        filename=os.path.join(HERE, "test_output", "test_full_run"),
        stype="science+standard",
        recreate=False,
        overwrite=True,
    )

    # Check the modified headers are still not overwritten
    print(
        lhs6328_onedspec.science_spectrum_list[0]
        .trace_hdulist[0]
        .header["COMMENT"]
    )

    # Create more FITS
    lhs6328_onedspec.create_fits(
        output="trace", stype="science+standard", recreate=True
    )

    # Now the FITS header should be back to default
    print(lhs6328_onedspec.science_spectrum_list[0].trace_hdulist[0].header)

    # save as CSV
    lhs6328_onedspec.save_csv(
        output="*",
        filename=os.path.join(HERE, "test_output", "test_full_run"),
        stype="science+standard",
        overwrite=True,
    )

    # save figure
    lhs6328_onedspec.inspect_reduced_spectrum(
        filename=os.path.join(HERE, "test_output", "test_full_run"),
        display=False,
        save_fig=True,
        fig_type="iframe+png",
    )

    # Plot the hough search space
    lhs6328_onedspec.plot_search_space(
        filename=os.path.join(
            HERE, "test_output", "test_full_run_plot_hough_space"
        ),
        display=False,
        save_fig=True,
        fig_type="iframe+png",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Functions for spectral extraction"""

import numpy as np
from scipy import ndimage, special
from statsmodels.nonparametric.smoothers_lowess import lowess

from .util import bfixpix, gaus


def tophat_extraction(
    source_slice: np.ndarray,
    sky_source_slice: np.ndarray,
    var_sky: float,
    pix_frac: float,
    gain: float,
    sky_width_dn: int,
    sky_width_up: int,
    width_dn: int,
    width_up: int,
    source_bad_mask: np.ndarray = None,
    sky_source_bad_mask: np.ndarray = None,
):
    """
    Make sure the counts are the number of photoelectrons or an equivalent
    detector unit, and not counts per second.

    Parameters
    ----------
    source_slice: 1-d numpy array (N)
        The counts along the profile for aperture extraction.
    sky_source_slice: 1-d numpy array (M)
        Count of the fitted sky along the pix, has to be the same
        length
    var_sky: float
        The variance of the sky_source_slice (measured, not fitted)
    pix_frac: float
        The decimal places of the centroid.
    gain: float
        Detector gain, in electrons per ADU
    sky_width_dn: int
        Number of pixels used for sky modelling on the lower side of the
        spectrum.
    sky_width_up: int
        Number of pixels used for sky modelling on the upper side of the
        spectrum.
    width_dn: int
        Number of pixels used for aperture extraction on the lower side
        of the spectrum.
    width_up: int
        Number of pixels used for aperture extraction on the upper side
        of the spectrum.
    source_bad_mask: 1-d numpy array (N, default: None)
        Masking the unusable pixels for extraction.
    sky_source_bad_mask: 1-d numpy array (M, default: None)
        Masking the unusable pixels for sky subtraction.

    """

    if source_bad_mask is not None:
        _source_slice = source_slice[~source_bad_mask]
    else:
        _source_slice = source_slice

    if source_bad_mask is not None:
        _sky_source_slice = sky_source_slice[~sky_source_bad_mask]
    else:
        _sky_source_slice = sky_source_slice

    # Get the total count
    source_plus_sky = (
        np.nansum(_source_slice)
        - pix_frac * _source_slice[0]
        - (1 - pix_frac) * _source_slice[-1]
    )

    # finally, compute the error in this pixel
    sky = (
        np.nansum(_sky_source_slice)
        - pix_frac * _sky_source_slice[0]
        - (1 - pix_frac) * _sky_source_slice[-1]
    )

    # number of bkgd pixels
    nB = sky_width_dn + sky_width_up - np.sum(np.isnan(_sky_source_slice))
    # number of aperture pixels
    nA = width_dn + width_up - np.sum(np.isnan(_source_slice))

    # Based on aperture phot err description by F. Masci,
    # Caltech:
    # http://wise2.ipac.caltech.edu/staff/fmasci/
    #   ApPhotUncert.pdf
    # All the counts are in per second already, so need to
    # multiply by the exposure time when computing the
    # uncertainty
    _signal = source_plus_sky - sky
    _noise = np.sqrt(_signal / gain + (nA + nA**2.0 / nB) * var_sky)

    return _signal, _noise, False


def optimal_extraction_horne86(
    source_slice: np.ndarray,
    sky: np.ndarray,
    profile: np.ndarray = None,
    tol: float = 1e-6,
    max_iter: int = 99,
    gain: float = 1.0,
    readnoise: float = 0.0,
    cosmicray_sigma: float = 5.0,
    forced: bool = False,
    variances: np.ndarray = None,
    bad_mask: np.ndarray = None,
):
    """
    Make sure the counts are the number of photoelectrons or an equivalent
    detector unit, and not counts per second or ADU.

    Iterate to get the optimal signal. Following the algorithm on
    Horne, 1986, PASP, 98, 609 (1986PASP...98..609H). The 'steps' in the
    inline comments are in reference to this article.

    The LOWESS setting can be found at:
    https://www.statsmodels.org/dev/generated/
        statsmodels.nonparametric.smoothers_lowess.lowess.html

    Parameters
    ----------
    source_slice: 1D numpy.ndarray (N)
        The counts along the profile for extraction, including the sky
        regions to be fitted and subtracted from. (NOT count per second)
    sky: 1D numpy.ndarray (N)
        Count of the fitted sky along the pix, has to be the same
        length as pix
        The width of the Gaussian
    profile: 1D numpy.ndarray (N)
        The gaussian profile (only used if model='gauss')
    tol: float
        The tolerance limit for the covergence
    max_iter: int
        The maximum number of iteration in the optimal extraction
    gain: float (Default: 1.0)
        Detector gain, in electrons per ADU
    readnoise: float
        Detector readnoise, in electrons.
    cosmicray_sigma: int (Default: 5)
        Sigma-clipping threshold for cleaning & cosmic ray rejection.
    forced: bool
        Forced extraction with the given weights.
    variances: 1D numpy.ndarray (N)
        The 1/weights of used for optimal extraction, has to be the
        same length as the pix. Only used if forced is True.
    bad_mask: list or None (Default: None)
        Mask of the bad or usable pixels.

    Returns
    -------
    signal: float
        The optimal signal.
    noise: float
        The noise associated with the optimal signal.
    is_optimal: bool
        List indicating whether the extraction at that pixel was
        optimal or not. True = optimal, False = suboptimal.
    P: numpy array
        The line spread function of the extraction
    var_f: float
        The variance in the extraction.

    """

    # step 2 - initial variance estimates
    var1 = readnoise**2.0 + np.abs(source_slice) / gain

    # step 4a - extract standard spectrum
    f = source_slice - sky
    f1 = np.nansum(f)

    # step 4b - variance of standard spectrum
    v1 = 1.0 / np.nansum(1.0 / var1)

    # step 5 - construct the spatial profile
    P = profile

    f_diff = 1
    v_diff = 1
    i = 0
    is_optimal = True

    while (f_diff > tol) | (v_diff > tol):
        mask_cr = np.ones(len(P), dtype=bool)

        if bad_mask is not None:
            mask_cr = mask_cr & ~bad_mask.astype(bool)

        if forced:
            var_f = variances

        f0 = f1
        v0 = v1

        # step 6 - revise variance estimates
        # var_f is the V in Horne87
        if not forced:
            var_f = readnoise**2.0 + np.abs(P * f0 + sky) / gain

        # step 7 - cosmic ray mask, only start considering after the
        # 2nd iteration. 1 pixel is masked at a time until convergence,
        # once the pixel is masked, it will stay masked.
        if i > 1:
            ratio = (cosmicray_sigma**2.0 * var_f) / (f - P * f0) ** 2.0

            if (ratio > 1).any():
                mask_cr[np.argmax(ratio)] = False

        denom = np.nansum((P**2.0 / var_f)[mask_cr])

        # step 8a - extract optimal signal
        f1 = np.nansum((P * f / var_f)[mask_cr]) / denom

        # step 8b - variance of optimal signal
        v1 = np.nansum(P[mask_cr]) / denom

        f_diff = abs((f1 - f0) / f0)
        v_diff = abs((v1 - v0) / v0)

        i += 1

        if i == int(max_iter):
            is_optimal = False
            break

    signal = f1
    noise = np.sqrt(v1)

    return signal, noise, is_optimal, P, var_f


def optimal_extraction_marsh89(
    frame,
    residual_frame: np.ndarray,
    variance: np.ndarray,
    trace: np.ndarray,
    spectrum: np.ndarray = None,
    readnoise: float = 0.0,
    apwidth: int = 7,
    goodpixelmask: np.ndarray = None,
    npoly: int = 21,
    polyspacing: int = 1,
    pord: int = 2,
    cosmicray_sigma: float = 5.0,
    qmode: str = "slow-nearest",
    nreject: int = 100,
):
    """
    Optimally extract curved spectra taken and updated from
    Ian Crossfield's code

    https://people.ucsc.edu/~ianc/python/_modules/spec.html#superExtract,
    following Marsh 1989.

    Parameters
    ----------
    frame: 2-d Numpy array (M, N)
        The calibrated frame from which to extract spectrum. In units
        of electrons count.
    residual_frame: 2-d Numpy array (M, N)
        The sky background only frame.
    variance: 2-d Numpy array (M, N)
        Variances of pixel values in 'frame'.
    trace: 1-d numpy array (N)
        :ocation of spectral trace.
    spectrum: 1-d numpy array (M) (Default: None)
        The extracted spectrum for initial guess.
    gain: float (Default: 1.0)
        Detector gain, in electrons per ADU
    readnoise: float (Default: 0.0)
        Detector readnoise, in electrons.
    apwidth: int or list of int (default: 7)
        The size of the aperture for extraction.
    goodpixelmask : 2-d numpy array (M, N) (Default: None)
        Equals 0 for bad pixels, 1 for good pixels
    npoly: int (Default: 21)
        Number of profile to be use for polynomial fitting to evaluate
        (Marsh's "K"). For symmetry, this should be odd.
    polyspacing: float (Default: 1)
        Spacing between profile polynomials, in pixels. (Marsh's "S").
        A few cursory tests suggests that the extraction precision
        (in the high S/N case) scales as S^-2 -- but the code slows down
        as S^2.
    pord: int (Default: 2)
        Order of profile polynomials; 1 = linear, etc.
    cosmicray_sigma: int (Default: 5)
        Sigma-clipping threshold for cleaning & cosmic-ray rejection.
    qmode: str (Default: 'slow-nearest')
        How to compute Marsh's Q-matrix. Valid inputs are 'fast-linear',
        'slow-linear', 'fast-nearest', and 'slow-nearest'. These select
        between various methods of integrating the nearest-neighbor or
        linear interpolation schemes as described by Marsh; the 'linear'
        methods are preferred for accuracy. Use 'slow' if you are
        running out of memory when using the 'fast' array-based methods.
    nreject: int (Default: 100)
        Number of outlier-pixels to reject at each iteration.

    Returns
    -------
    spectrum_marsh:
        The optimal signal.
    spectrum_err_marsh:
        The noise associated with the optimal signal.
    is_optimal:
        List indicating whether the extraction at that pixel was
        optimal or not (this list is always all optimal).
    profile:
        The line spread functions of the extraction
    variance0:
        The variance in the extraction.

    """

    frame = frame.transpose()
    residual_frame = residual_frame.transpose()
    variance = variance.transpose()

    if isinstance(apwidth, (float, int)):
        # first do the aperture count
        width_dn = apwidth
        width_up = apwidth

    elif isinstance(apwidth, (list, np.ndarray)) & (len(apwidth) == 2):
        width_dn = apwidth[0]
        width_up = apwidth[1]

    else:
        width_dn = 7
        width_up = 7

    if goodpixelmask is not None:
        goodpixelmask = goodpixelmask.transpose()
        goodpixelmask = np.array(goodpixelmask, copy=True).astype(bool)

    else:
        goodpixelmask = np.ones_like(frame, dtype=bool)

    goodpixelmask *= np.isfinite(frame) * np.isfinite(variance)

    variance[~goodpixelmask] = frame[goodpixelmask].max() * 1e9
    spectral_size, spatial_size = frame.shape

    # (my 3a: mask any bad values)
    bad_residual_frame_mask = ~np.isfinite(residual_frame)
    residual_frame[bad_residual_frame_mask] = 0.0

    sky_subframe = frame - residual_frame
    # Interpolate and fix bad pixels for extraction of standard
    # spectrum -- otherwise there can be 'holes' in the spectrum from
    # ill-placed bad pixels.
    sky_subframe = bfixpix(
        sky_subframe, ~goodpixelmask, n_nearest=8, retdat=True
    )

    # Define new indices (in Marsh's appendix):
    N = pord + 1
    mm = np.tile(np.arange(N).reshape(N, 1), (npoly)).ravel()
    nn = mm.copy()
    ll = np.tile(np.arange(npoly), N)
    kk = ll.copy()
    pp = N * ll + mm
    qq = N * kk + nn

    ii = np.arange(spatial_size)  # column (i.e., spatial direction)
    jjnorm = np.linspace(-1, 1, spectral_size)  # normalized X-coordinate
    _pow = np.arange(2 * N - 1).reshape(2 * N - 1, 1, 1)
    jjnorm_pow = jjnorm.reshape(1, 1, spectral_size) ** _pow

    # Marsh eq. 9, defining centers of each polynomial:
    constant = 0.0  # What is it for???
    poly_centers = (
        np.array(trace).reshape(spectral_size, 1)
        + polyspacing * np.arange(-npoly / 2 + 1, npoly / 2 + 1)
        + constant
    )

    # Marsh eq. 11, defining Q_kij    (via nearest-neighbor interpolation)
    #    Q_kij =  max(0, min(S, (S+1)/2 - abs(x_kj - i)))
    if qmode == "fast-nearest":  # Array-based nearest-neighbor mode.
        Q = np.array(
            [
                np.zeros((npoly, spatial_size, spectral_size)),
                np.array(
                    [
                        polyspacing
                        * np.ones((npoly, spatial_size, spectral_size)),
                        0.5 * (polyspacing + 1)
                        - np.abs(
                            (
                                poly_centers - ii.reshape(spatial_size, 1, 1)
                            ).transpose(2, 0, 1)
                        ),
                    ]
                ).min(0),
            ]
        ).max(0)

    elif qmode == "slow-linear":  # Code is a mess, but it works.
        invs = 1.0 / polyspacing
        poly_centers_over_s = poly_centers / polyspacing
        xps_mat = poly_centers + polyspacing
        xms_mat = poly_centers - polyspacing
        Q = np.zeros((npoly, spatial_size, spectral_size))
        for i in range(spatial_size):
            ip05 = i + 0.5
            im05 = i - 0.5
            for j in range(spectral_size):
                for k in range(npoly):
                    xkj = poly_centers[j, k]
                    xkjs = poly_centers_over_s[j, k]
                    # xkj + polyspacing
                    xps = xps_mat[j, k]
                    # xkj - polyspacing
                    xms = xms_mat[j, k]

                    if (ip05 <= xms) or (im05 >= xps):
                        qval = 0.0
                    elif (im05) > xkj:
                        lim1 = im05
                        lim2 = min(ip05, xps)
                        qval = (lim2 - lim1) * (
                            1.0 + xkjs - 0.5 * invs * (lim1 + lim2)
                        )
                    elif (ip05) < xkj:
                        lim1 = max(im05, xms)
                        lim2 = ip05
                        qval = (lim2 - lim1) * (
                            1.0 - xkjs + 0.5 * invs * (lim1 + lim2)
                        )
                    else:
                        lim1 = max(im05, xms)
                        lim2 = min(ip05, xps)
                        qval = (
                            lim2
                            - lim1
                            + invs
                            * (
                                xkj * (-xkj + lim1 + lim2)
                                - 0.5 * (lim1 * lim1 + lim2 * lim2)
                            )
                        )
                    Q[k, i, j] = max(0, qval)

    # Code is a mess, but it's faster than 'slow' mode
    elif qmode == "fast-linear":
        invs = 1.0 / polyspacing
        xps_mat = poly_centers + polyspacing
        Q = np.zeros((npoly, spatial_size, spectral_size))
        for j in range(spectral_size):
            xkj_vec = np.tile(
                poly_centers[j, :].reshape(npoly, 1), (1, spatial_size)
            )
            xps_vec = np.tile(
                xps_mat[j, :].reshape(npoly, 1), (1, spatial_size)
            )
            xms_vec = xps_vec - 2 * polyspacing

            ip05_vec = np.tile(np.arange(spatial_size) + 0.5, (npoly, 1))
            im05_vec = ip05_vec - 1
            ind00 = (ip05_vec <= xms_vec) + (im05_vec >= xps_vec)
            ind11 = (im05_vec > xkj_vec) * ~ind00
            ind22 = (ip05_vec < xkj_vec) * ~ind00
            ind33 = ~(ind00 + ind11 + ind22)
            ind11 = ind11.nonzero()
            ind22 = ind22.nonzero()
            ind33 = ind33.nonzero()

            n_ind11 = len(ind11[0])
            n_ind22 = len(ind22[0])
            n_ind33 = len(ind33[0])

            if n_ind11 > 0:
                ind11_3d = ind11 + (np.ones(n_ind11, dtype=int) * j,)
                lim2_ind11 = np.array((ip05_vec[ind11], xps_vec[ind11])).min(0)
                Q[ind11_3d] = (
                    (lim2_ind11 - im05_vec[ind11])
                    * invs
                    * (
                        polyspacing
                        + xkj_vec[ind11]
                        - 0.5 * (im05_vec[ind11] + lim2_ind11)
                    )
                )

            if n_ind22 > 0:
                ind22_3d = ind22 + (np.ones(n_ind22, dtype=int) * j,)
                lim1_ind22 = np.array((im05_vec[ind22], xms_vec[ind22])).max(0)
                Q[ind22_3d] = (
                    (ip05_vec[ind22] - lim1_ind22)
                    * invs
                    * (
                        polyspacing
                        - xkj_vec[ind22]
                        + 0.5 * (ip05_vec[ind22] + lim1_ind22)
                    )
                )

            if n_ind33 > 0:
                ind33_3d = ind33 + (np.ones(n_ind33, dtype=int) * j,)
                lim1_ind33 = np.array((im05_vec[ind33], xms_vec[ind33])).max(0)
                lim2_ind33 = np.array((ip05_vec[ind33], xps_vec[ind33])).min(0)
                Q[ind33_3d] = (lim2_ind33 - lim1_ind33) + invs * (
                    xkj_vec[ind33]
                    * (-xkj_vec[ind33] + lim1_ind33 + lim2_ind33)
                    - 0.5 * (lim1_ind33 * lim1_ind33 + lim2_ind33 * lim2_ind33)
                )

    # 'slow' Loop-based nearest-neighbor mode: requires less memory
    else:
        Q = np.zeros((npoly, spatial_size, spectral_size))
        for k in range(npoly):
            for i in range(spatial_size):
                for j in range(spectral_size):
                    Q[k, i, j] = max(
                        0,
                        min(
                            polyspacing,
                            0.5 * (polyspacing + 1)
                            - np.abs(poly_centers[j, k] - i),
                        ),
                    )

    # Some quick math to find out which dat columns are important, and
    # which contain no useful spectral information:
    q_mask = Q.sum(0).transpose() > 0
    q_ind = q_mask.transpose().nonzero()
    q_cols = [q_ind[0].min(), q_ind[0].max()]
    q_sm = Q[:, q_cols[0] : q_cols[1] + 1, :]

    # Prepar to iteratively clip outliers
    new_bad_pixels = True
    i = -1

    spectrum_marsh = np.zeros(spectral_size)
    spectrum_err_marsh = np.zeros(spectral_size)
    is_optimal = np.zeros(spectral_size)
    profile = np.zeros((spatial_size, spectral_size))
    model_spectrum = (
        np.array(spectrum).reshape(spectral_size, 1) * profile.transpose()
    )
    model_data = model_spectrum + residual_frame
    variance0 = np.abs(model_data) + readnoise**2

    while new_bad_pixels:
        i += 1

        # Compute pixel fractions (Marsh Eq. 5):
        #     (Note that values outside the desired polynomial region
        #     have Q=0, and so do not contribute to the fit)
        inv_e_variance = (
            np.array(spectrum).reshape(spectral_size, 1) ** 2 / variance
        ).transpose()
        weighted_e = (
            sky_subframe
            * np.array(spectrum).reshape(spectral_size, 1)
            / variance
        ).transpose()  # E / var_E
        inv_e_variance_subset = inv_e_variance[q_cols[0] : q_cols[1] + 1, :]

        # Define X vector (Marsh Eq. A3):
        X = np.zeros(N * npoly)
        for q in qq:
            X[q] = (
                weighted_e[q_cols[0] : q_cols[1] + 1, :]
                * q_sm[kk[q], :, :]
                * jjnorm_pow[nn[q]]
            ).sum()

        # Define C matrix (Marsh Eq. A3)
        C = np.zeros((N * npoly, N * npoly))

        # C-matrix computation buffer (to be sure we don't miss any pixels)
        buffer = 1.1

        # Compute *every* element of C (though most equal zero!)
        for p in pp:
            qp = q_sm[ll[p], :, :]
            for q in qq:
                #  Check that we need to compute C:
                if np.abs(kk[q] - ll[p]) <= (1.0 / polyspacing + buffer):
                    if q >= p:
                        # Only compute over non-zero columns:
                        C[q, p] = (
                            q_sm[kk[q], :, :]
                            * qp
                            * jjnorm_pow[nn[q] + mm[p]]
                            * inv_e_variance_subset
                        ).sum()
                    if q > p:
                        C[p, q] = C[q, p]

        # Solve for the profile-polynomial coefficients (Marsh Eq. A4):
        if np.abs(np.linalg.det(C)) < 1e-10:
            b_soln = np.dot(np.linalg.pinv(C), X)
        else:
            b_soln = np.linalg.solve(C, X)

        a_soln = b_soln.reshape(N, npoly).transpose()

        # Define G_kj, the profile-defining polynomial profiles
        # (Marsh Eq. 8)
        g_soln = np.zeros((npoly, spectral_size))
        for n in range(npoly):
            g_soln[n] = np.polyval(
                a_soln[n, ::-1], jjnorm
            )  # reorder polynomial coef.

        # Compute the profile (Marsh eq. 6) and normalize it:
        for i in range(spatial_size):
            profile[i, :] = (Q[:, i, :] * g_soln).sum(0)

        if profile.min() < 0:
            profile[profile < 0] = 0.0
        profile /= np.nansum(profile, axis=0)
        profile[~np.isfinite(profile)] = 0.0

        # Step6: Revise variance estimates
        model_spectrum = (
            np.array(spectrum).reshape(spectral_size, 1) * profile.transpose()
        )
        model_data = model_spectrum + residual_frame
        variance0 = np.abs(model_data) + readnoise**2
        variance = variance0 / (
            goodpixelmask + 1e-9
        )  # De-weight bad pixels, avoiding infinite variance

        outlier_variances = (frame - model_data) ** 2 / variance

        if outlier_variances.max() > cosmicray_sigma**2:
            new_bad_pixels = True
            # nreject-counting on pixels within the spectral trace
            max_rejected_value = max(
                cosmicray_sigma**2,
                np.sort(outlier_variances[q_mask])[-nreject],
            )
            worst_outliers = (
                outlier_variances >= max_rejected_value
            ).nonzero()
            goodpixelmask[worst_outliers] = False
            # number_rejected = len(worst_outliers[0])
        else:
            new_bad_pixels = False
            # number_rejected = 0

        # Optimal Spectral Extraction: (Horne, Step 8)
        for i in range(spectral_size):
            aperture = np.arange(
                int(trace[i]) - width_dn, int(trace[i]) + width_up + 1
            ).astype(int)

            # Horne86 notation
            P = profile[aperture, i]
            V = variance0[i, aperture]
            D = sky_subframe[i, aperture]

            denom = np.nansum(P**2.0 / V)

            if denom == 0:
                spectrum_marsh[i] = 0.0
                spectrum_err_marsh[i] = 9e9
            else:
                spectrum_marsh[i] = np.nansum(P / V * D) / denom
                spectrum_err_marsh[i] = np.sqrt(np.nansum(P) / denom)
                is_optimal[i] = True

    return (
        spectrum_marsh,
        spectrum_err_marsh,
        is_optimal,
        profile.T,
        variance0,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""ASPIRED"""

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass  # package is not installed

__status__ = "Production"
__credits__ = [
    "Marco C Lam",
    "Robert J Smith",
    "Iair Arcavi",
    "Iain A Steele",
    "Josh Veitch-Michaelis",
    "Lukasz Wyrzykowski",
]

from . import image_reduction, spectral_reduction, standard_list

__all__ = [
    "image_reduction",
    "spectral_reduction",
    "standard_list",
    "util",
    "flux_calibration",
    "wavelength_calibration",
]



# -*- coding: utf-8 -*-
import copy
import os
from unittest.mock import patch

import numpy as np
from aspired import spectral_reduction
from aspired.flux_calibration import FluxCalibration

HERE = os.path.dirname(os.path.realpath(__file__))


def test_telluric_square_wave():
    wave = np.arange(1000.0)
    flux_sci = np.ones(1000) * 5.0
    flux_std = np.ones(1000) * 100.0

    flux_sci_continuum = copy.deepcopy(flux_sci)
    flux_std_continuum = copy.deepcopy(flux_std)

    flux_sci[500:550] *= 0.01
    flux_sci[700:750] *= 0.001
    flux_sci[850:950] *= 0.1

    flux_std[500:550] *= 0.01
    flux_std[700:750] *= 0.001
    flux_std[850:950] *= 0.1

    # Get the telluric profile
    fluxcal = FluxCalibration(log_file_name=None)
    telluric_func, _, _ = fluxcal.get_telluric_profile(
        wave,
        flux_std,
        flux_std_continuum,
        mask_range=[[495, 551], [700, 753], [848, 960]],
        return_function=True,
    )

    onedspec = spectral_reduction.OneDSpec(log_file_name=None)
    onedspec.science_spectrum_list[0].add_wavelength(wave)
    onedspec.science_spectrum_list[0].add_flux(flux_sci, None, None)
    onedspec.science_spectrum_list[0].add_flux_continuum(flux_sci_continuum)
    onedspec.science_data_available = True
    # onedspec.fluxcal.spectrum_oned.add_wavelength(wave)
    # onedspec.fluxcal.spectrum_oned.add_flux(flux_std, None, None)
    # onedspec.fluxcal.spectrum_oned.add_flux_continuum(flux_std_continuum)

    onedspec.add_telluric_function(telluric_func, stype="science")
    onedspec.get_telluric_strength()
    onedspec.apply_telluric_correction()

    assert np.isclose(
        np.nansum(onedspec.science_spectrum_list[0].flux_telluric_corrected),
        np.nansum(flux_sci_continuum),
        rtol=1e-2,
    )

    onedspec.inspect_telluric_correction(
        display=False,
        return_jsonstring=True,
        save_fig=True,
        fig_type="iframe+jpg+png+svg+pdf",
        filename=os.path.join(HERE, "test_output", "test_telluric"),
    )


@patch("plotly.graph_objects.Figure.show")
def test_telluric_real_data(mock_show):
    std_wave = np.load(os.path.join(HERE, "test_data", "std_wave.npy"))
    std_flux = np.load(os.path.join(HERE, "test_data", "std_flux.npy"))
    std_flux_continuum = np.load(
        os.path.join(HERE, "test_data", "std_flux_continuum.npy")
    )
    sci_wave = np.load(os.path.join(HERE, "test_data", "sci_wave.npy"))
    sci_flux = np.load(os.path.join(HERE, "test_data", "sci_flux.npy"))
    sci_flux_continuum = np.load(
        os.path.join(HERE, "test_data", "sci_flux_continuum.npy")
    )

    # Get the telluric profile
    fluxcal = FluxCalibration(log_file_name=None)
    telluric_func, _, _ = fluxcal.get_telluric_profile(
        std_wave, std_flux, std_flux_continuum, return_function=True
    )

    onedspec = spectral_reduction.OneDSpec(log_file_name=None)
    onedspec.science_spectrum_list[0].add_wavelength(sci_wave)
    onedspec.science_spectrum_list[0].add_flux(sci_flux, None, None)
    onedspec.science_spectrum_list[0].add_flux_continuum(sci_flux_continuum)
    onedspec.science_data_available = True
    onedspec.fluxcal.spectrum_oned.add_wavelength(std_wave)
    onedspec.fluxcal.spectrum_oned.add_flux(std_flux, None, None)
    onedspec.fluxcal.spectrum_oned.add_flux_continuum(std_flux_continuum)

    onedspec.add_telluric_function(telluric_func)
    onedspec.apply_telluric_correction()

    assert np.isclose(
        np.nansum(onedspec.science_spectrum_list[0].flux),
        np.nansum(sci_flux_continuum),
        rtol=1e-2,
    )

    onedspec.inspect_telluric_profile(
        display=True,
        return_jsonstring=True,
        save_fig=True,
        fig_type="iframe+jpg+png+svg+pdf",
        filename=os.path.join(HERE, "test_output", "test_telluric"),
    )



# -*- coding: utf-8 -*-
import glob
import json
import os
import sys

try:
    base_dir = os.path.abspath(os.path.dirname(__file__))
except:
    base_dir = os.path.abspath(os.path.dirname(__name__))

sys.path.insert(0, os.path.join(base_dir, ".."))

from aspired.standard_list import library_list
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.simbad import Simbad

# ignore for now
# (1) a dictionary to match names to unique names (the search is one-to-one)
# (2) a dictionary to unique names to names (the search is one-to-many)

# essential
# (3) a dictionary to match libraries to unique names (the search is one-to-many)
# (4) a dictionary to match unique names to libraries (the search is one-to-many)

# the order in which the standard libraries goes through in automated mode
library_rank = {
    "ingoke": 1,
    "ingmas": 2,
    "irafctionewcal": 3,
    "irafspec16cal": 4,
    "ingog": 5,
    "esoxshooter": 6,
    "esowdstan": 7,
    "esohststan": 8,
    "esookestan": 9,
    "esoctiostan": 10,
    "irafctiocal": 11,
    "irafiidscal": 12,
    "irafredcal": 13,
    "irafbstdscal": 14,
    "irafirscal": 15,
    "ingsto": 16,
    "irafspec50cal": 17,
    "irafoke1990": 18,
    "irafspechayescal": 19,
    "ingfg": 20,
    "irafblackbody": 21,
}


filelist = glob.glob("*filename-to-identifier.json")


# Create dictionary for libraries to filenames
lib_to_filename = {}
lib_to_designation = {}

# And the mapping of designations and filename
lib_filename_to_designation_position = {}
designation_to_lib_filename = {}
designation_to_position = {}

for json_filename in filelist:
    data = json.load(open(json_filename, encoding="ascii"))
    lib_name = json_filename.split("-")[0]
    lib_to_filename[lib_name] = list(data.keys())
    for k, v in data.items():
        if lib_name == "irafblackbody":
            ra = 0.0
            dec = 0.0
        else:
            _ra, _dec = Simbad.query_object(v[0])["RA", "DEC"].items()
            c = SkyCoord(
                _ra[-1].data.data[0] + " " + _dec[-1].data.data[0],
                unit=(u.hourangle, u.deg),
            )
            ra = c.ra.value
            dec = c.dec.value
        lib_to_designation[lib_name] = [d.replace(" ", "").lower() for d in v]
        lib_filename_to_designation_position[lib_name + "/" + k] = [
            [d.replace(" ", "").lower(), ra, dec] for d in v
        ]


# Loop through the library names
for (
    lib_filename,
    designation_position,
) in lib_filename_to_designation_position.items():
    libname, filename = lib_filename.split("/")
    for d, _r, _d in designation_position:
        if d.lower() not in designation_to_lib_filename:
            designation_to_lib_filename[d.lower()] = {}
            designation_to_position[d.lower()] = {}
        designation_to_lib_filename[d.lower()][libname] = filename
        designation_to_position[d.lower()][libname] = [_r, _d]

# Create dictionary for filenames to libraries
filename_to_lib = {}

# Loop through the library names
for lib, stars in lib_to_filename.items():
    for s in stars:
        if s.lower() not in filename_to_lib:
            filename_to_lib[s.lower()] = [lib.lower()]
        else:
            filename_to_lib[s.lower()].append(lib.lower())

# Sort the list of the dictionaries
for star, libs in filename_to_lib.items():
    filename_to_lib[star] = sorted(
        filename_to_lib[star], key=lambda i: library_rank[i]
    )

with open("lib_to_filename.json", "w") as f:
    json.dump(lib_to_filename, f)

with open("lib_to_designation.json", "w") as f:
    json.dump(lib_to_designation, f)

with open("filename_to_lib.json", "w") as f:
    json.dump(filename_to_lib, f)

with open("lib_filename_to_designation_position.json", "w") as f:
    json.dump(lib_filename_to_designation_position, f)

with open("designation_to_lib_filename.json", "w") as f:
    json.dump(designation_to_lib_filename, f)

with open("designation_to_position.json", "w") as f:
    json.dump(designation_to_position, f)
